#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level directory facilities.
'''

# ....................{ IMPORTS                            }....................
import os, shutil
from betse.exceptions import BetseDirException, BetsePathException
from betse.util.io.log import logs
from betse.util.type.enums import make_enum
from betse.util.type.types import (
    type_check,
    CallableTypes,
    GeneratorType,
    IterableOrNoneTypes,
    NumericSimpleTypes,
    SequenceTypes,
)
from distutils import dir_util
from os import path as os_path

# ....................{ GLOBALS ~ enum                     }....................
DirOverwritePolicy = make_enum(
    class_name='DirOverwritePolicy',
    member_names=(
        'HALT_WITH_EXCEPTION',
        'SKIP_WITH_WARNING',
        'OVERWRITE',
    ),
)
'''
Enumeration of all supported types of **directory overwrite policies** (i.e.,
strategies for handling existing paths visited by recursive directory
operations).

Attributes
----------
HALT_WITH_EXCEPTION : enum
    Policy raising a fatal exception if any target path already exists. This
    constitutes the strictest and hence safest such policy.
SKIP_WITH_WARNING : enum
    Policy ignoring (i.e., skipping) each target path that already exists with a
    logged non-fatal warning. This policy strikes a convenient balance between
    strictness and laxness.
OVERWRITE : enum
    Policy silently overwriting each target path that already exists. This
    constitutes the laxest and hence riskiest such policy.
'''

# ....................{ GLOBALS ~ regex                    }....................
# There exist only two possible directory separators for all modern platforms.
# Hence, this reliably suffices with no error handling required.
SEPARATOR_REGEX = r'/' if os_path.sep == '/' else r'\\'
'''
Regular expression matching the directory separator specific to the current
platform.

Specifically, under:

* Microsoft Windows, this is `\\\\`.
* All other platforms, this is `/`.
'''

# ....................{ EXCEPTIONS                         }....................
def die_if_dir(*dirnames: str) -> None:
    '''
    Raise an exception if any of the passed directories exist.
    '''

    for dirname in dirnames:
        if is_dir(dirname):
            raise BetseDirException(
                'Directory "{}" already exists.'.format(dirname))


def die_unless_dir(*dirnames: str) -> None:
    '''
    Raise an exception unless all passed directories exist.
    '''

    for dirname in dirnames:
        if not is_dir(dirname):
            raise BetseDirException(
                'Directory "{}" not found or unreadable.'.format(dirname))


def join_and_die_unless_dir(*pathnames: str) -> str:
    '''
    Pathname of the directory produced by joining (i.e., concatenating) the
    passed pathnames with the directory separator specific to the current
    platform if this directory exists *or* raise an exception otherwise (i.e.,
    if this directory does *not* exist).

    This higher-level function is a convenience wrapper encapsulating both the
    lower-level :func:`betse.util.path.pathnames.join` and
    :func:`die_unless_dir` functions.
    '''

    # Avoid circular import dependencies.
    from betse.util.path.pathnames import join

    # Dirname producing by joining these pathnames.
    dirname = join(*pathnames)

    # If this directory does *NOT* exist, raise an exception.
    die_unless_dir(dirname)

    # Return this dirname.
    return dirname

# ....................{ EXCEPTIONS ~ parent                }....................
def die_unless_parent_dir(pathname: str) -> None:
    '''
    Raise an exception unless the parent directory of the passed path exists.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import pathnames

    # If the parent directory of this path does *NOT* exist, raise an exception.
    die_unless_dir(pathnames.get_dirname(pathname))

# ....................{ TESTERS                            }....................
@type_check
def is_dir(dirname: str) -> bool:
    '''
    ``True`` only if the directory with the passed path exists.
    '''

    return os_path.isdir(dirname)

# ....................{ GETTERS                            }....................
@type_check
def get_parent_dir_last(pathname: str) -> str:
    '''
    Absolute pathname of the most deeply nested existing parent directory of the
    path with the passed pathname.

    Since the passed pathname is required to be absolute *and* since the root
    directory (e.g., ``/`` on POSIX-compatible platforms) always exists, this
    function is guaranteed to return the pathname of an existing directory.

    Parameters
    -----------
    pathname : str
        Absolute pathname of the path to inspect.

    Returns
    -----------
    str
        Absolute pathname of the most deeply nested existing parent directory of
        this path.

    Examples
    -----------
    Given an existing directory ``/the/garden/of/earthly/`` and non-existing
    subdirectory ``/the/garden/of/earthly/delights/`` of that directory:

        >>> from betse.util.path import dirs
        >>> dirs.get_parent_dir_last('/the/garden/of/earthly/delights/')
        '/the/garden/of/earthly'
    '''

    # Avoid circular import dependencies.
    from betse.util.path import pathnames

    # If this pathname is relative rather than absolute, raise an exception.
    pathnames.die_if_relative(pathname)

    # Current directory component of this path to inspect the existence of.
    dirname_component = pathname

    # While this directory component is *NOT* an existing directory...
    while not is_dir(dirname_component):
        # Reduce this directory component to its parent directory.
        dirname_component = pathnames.get_dirname(dirname_component)

    # Return this existing directory component.
    return dirname_component

# ....................{ GETTERS ~ mtime : recursive        }....................
#FIXME: Consider contributing the get_mtime_recursive_newest() function as an
#answer to the following StackOverflow question:
#    https://stackoverflow.com/questions/26498285/find-last-update-time-of-a-directory-includes-all-levels-of-sub-folders

# If the current platform provides the os.fwalk() function *AND* the os.stat()
# function implemented by this platform accepts directory handles (which is
# currently equivalent to testing if this platform is POSIX-compatible),
# optimize this recursion by leveraging these handles.
if hasattr(os, 'fwalk') and os.stat in os.supports_dir_fd:
    @type_check
    def get_mtime_recursive_newest(dirname: str) -> NumericSimpleTypes:

        # Log this recursion.
        logs.log_debug(
            'Recursively computing newest fwalk()-based mtime of: %s', dirname)

        # Return the maximum of all mtimes in the...
        return max(
            # Generator expression recursively yielding the maximum mtime of all
            # files and subdirectories of each subdirectory of this directory
            # including this directory.
            (
                # Maximum of this subdirectory's mtime and the mtimes of all
                # files in this subdirectory, whose filenames are produced by a
                # tuple comprehension joining this subdirectory's dirname to
                # each file's basename. By recursion, the mtimes of all
                # subdirectories of this subdirectory are implicitly computed
                # and hence need *NOT* be explicitly included here.
                #
                # For parity with the unoptimized get_mtime_recursive_newest()
                # implementation defined below as well to avoid unwanted
                # complications, symbolic links are *NOT* followed. Hence,
                # os.lstat() rather than os.stat() is intentionally called.
                max(
                    (os.fstat(parent_dir_fd).st_mtime,) + tuple(
                        os.lstat(
                            child_file_basename, dir_fd=parent_dir_fd).st_mtime
                        for child_file_basename in child_file_basenames
                    ),
                )
                # For the currently visited directory's dirname, a sequence of
                # the dirnames of all subdirectories of this directory, a
                # sequence of the filenames of all files in this directory, and
                # a directory handle to this directory such that errors emitted
                # by low-level functions called by os.walk() (e.g.,
                # os.listdir()) are *NOT* silently ignored...
                for _, _, child_file_basenames, parent_dir_fd in _fwalk(dirname)
            ),
        )

# Else, fallback to unoptimized recursion leveraging dirnames.
else:
    @type_check
    def get_mtime_recursive_newest(dirname: str) -> NumericSimpleTypes:

        # Log this recursion.
        logs.log_debug(
            'Recursively computing newest walk()-based mtime of: %s', dirname)

        # Return the maximum of all mtimes in the...
        return max(
            # Generator expression recursively yielding the maximum mtime of all
            # files and subdirectories of each subdirectory of this directory
            # including this directory.
            (
                # Maximum of this subdirectory's mtime and the mtimes of all
                # files in this subdirectory, whose filenames are produced by a
                # tuple comprehension joining this subdirectory's dirname to
                # each file's basename. By recursion, the mtimes of all
                # subdirectories of this subdirectory are implicitly computed
                # and hence need *NOT* be explicitly included here.
                max(
                    (os_path.getmtime(parent_dirname),) + tuple(
                        os_path.getmtime(os_path.join(
                            parent_dirname, child_file_basename))
                        for child_file_basename in child_file_basenames
                    ),
                )
                # For the currently visited directory's dirname, a sequence of
                # the dirnames of all subdirectories of this directory, and a
                # sequence of the filenames of all files in this directory such
                # that errors emitted by low-level functions called by
                # os.walk() (e.g., os.listdir()) are *NOT* silently ignored...
                for parent_dirname, _, child_file_basenames in _walk(dirname)
            ),
        )

# Docstring dynamically set for the getter defined above.
get_mtime_recursive_newest.__doc__ = '''
    Most recent **recursive mtime** (i.e., recursively calculated modification
    time) in seconds of all paths in the set of this directory and all files and
    subdirectories transitively reachable from this directory *witthout*
    following symbolic links to directories.

    Caveats
    -----------
    Note that symbolic links whose transitive targets are:

    * Directories are *not* followed, avoiding infinite recursion in edge cases
      (e.g., directories containing symbolic links to themselves). Such links
      are effectively non-directory files. Since the mtime of a symbolic link is
      typically its ctime (i.e., creation time), such links thus contribute
      their own ctime to this calculation.
    * Files are followed, improving the accuracy of this calculation.

    Parameters
    -----------
    dirname : str
        Absolute or relative path of the directory to inspect.

    Returns
    -----------
    NumericSimpleTypes
        Most recent mtime in seconds of this directory calculated recursively as
        either an integer or float, depending on the boolean returned by the
        platform-specific :func:`os.stat_float_times` function.

    Raises
    -----------
    OSError
        If either this directory *or* any file or subdirectory reachable from
        this directory is unreadable by the current user (e.g., due to
        insufficient permissions).
    '''

# ....................{ COPIERS                            }....................
@type_check
def copy_into_dir(src_dirname: str, trg_dirname: str, *args, **kwargs) -> None:
    '''
    Recursively copy the passed source directory to a subdirectory of the passed
    target directory having the same basename as this source directory.

    Parameters
    -----------
    src_dirname : str
        Absolute or relative path of the source directory to be recursively
        copied from.
    trg_dirname : str
        Absolute or relative path of the target directory to copy into.

    All remaining parameters are passed as is to the :func:`copy` function.

    See Also
    ----------
    :func:`copy`
        Further details.

    Examples
    ----------
        >>> from betse.util.path import dirs
        >>> dirs.copy_into_dir('/usr/src/linux/', '/tmp/')
        >>> dirs.is_dir('/tmp/linux/')
        True
    '''

    # Avoid circular import dependencies.
    from betse.util.path import pathnames

    # Basename of this source directory.
    src_basename = pathnames.get_basename(src_dirname)

    # Absolute or relative path of the target directory to copy to.
    trg_subdirname = pathnames.join(trg_dirname, src_basename)

    # Recursively copy this source to target directory.
    copy(*args, src_dirname=src_dirname, trg_dirname=trg_subdirname, **kwargs)


@type_check
def copy(
    # Mandatory parameters.
    src_dirname: str,
    trg_dirname: str,

    # Optional parameters.
    overwrite_policy: DirOverwritePolicy = (
        DirOverwritePolicy.HALT_WITH_EXCEPTION),
    ignore_basename_globs: IterableOrNoneTypes = None,
) -> None:
    '''
    Recursively copy the passed source to target directory.

    All nonexistent parents of the target directory will be recursively created,
    mimicking the action of the ``mkdir -p`` shell command. All symbolic links
    in the source directory will be preserved (i.e., copied as is rather than
    their transitive targets copied instead).

    Parameters
    -----------
    src_dirname : str
        Absolute or relative path of the source directory to be recursively
        copied from.
    trg_dirname : str
        Absolute or relative path of the target directory to recursively copy to.
    overwrite_policy : DirOverwritePolicy
        **Directory overwrite policy** (i.e., strategy for handling existing
        paths to be overwritten by this copy) to apply. Defaults to
        :attr:`DirOverwritePolicy.HALT_WITH_EXCEPTION`, raising an exception if
        any target path already exists.
    ignore_basename_globs : optional[IterableTypes]
        Iterable of shell-style globs (e.g., ``('*.tmp', '.keep')``) matching
        the basenames of all paths transitively owned by this source directory
        to be ignored during recursion and hence neither copied nor visited.
        If non-``None`` and the ``overwrite_policy`` parameter is
        :attr:`DirOverwritePolicy.OVERWRITE`, this iterable is ignored and a
        non-fatal warning is logged. Defaults to ``None``, in which case *all*
        paths transitively owned by this source directory are unconditionally
        copied and visited.

    Raises
    -----------
    BetseDirException
        If:
        * The source directory does not exist.
        * One or more subdirectories of the target directory already exist that
          are also subdirectories of the source directory. For safety, this
          function always preserves rather than overwrites existing target
          subdirectories.

    See Also
    -----------
    https://stackoverflow.com/a/22588775/2809027
        StackOverflow answer strongly inspiring this function's
        :attr:`DirOverwritePolicy.SKIP_WITH_WARNING` implementation.
    '''

    # Log this copy.
    logs.log_debug('Copying directory: %s -> %s', src_dirname, trg_dirname)

    # Raise an exception unless the source directory exists.
    die_unless_dir(src_dirname)

    # If passed an iterable of shell-style globs matching ignorable basenames,
    # convert this iterable into a predicate function of the form required by
    # the shutil.copytree() function. Specifically, this function accepts the
    # absolute or relative pathname of an arbitrary directory and an iterable of
    # the basenames of all subdirectories and files directly in this directory;
    # this function returns an iterable of the basenames of all subdirectories
    # and files in this directory to be ignored. This signature resembles:
    #
    #     def ignore_basename_func(
    #         parent_dirname: str,
    #         child_basenames: IterableTypes) -> IterableTypes
    ignore_basename_func = None
    if ignore_basename_globs is not None:
        ignore_basename_func = shutil.ignore_patterns(*ignore_basename_globs)

    # If raising a fatal exception if any target path already exists...
    if overwrite_policy is DirOverwritePolicy.HALT_WITH_EXCEPTION:
        # Dictionary of all keyword arguments to pass to shutil.copytree(),
        # preserving symbolic links as is.
        copytree_kwargs = {
            'symlinks': True,
        }

        # If ignoring basenames, inform shutil.copytree() of these basenames.
        if ignore_basename_func is not None:
            copytree_kwargs['ignore'] = ignore_basename_func

        # Raise an exception if this target directory already exists. While we
        # could defer to the exception raised by the shutil.copytree() function
        # for this case, this exception's message erroneously refers to this
        # directory as a file and is hence best avoided as non-human-readable:
        #
        #     [Errno 17] File exists: 'sample_sim'
        die_if_dir(trg_dirname)

        # Recursively copy this source to target directory. To avoid silently
        # overwriting all conflicting target paths, the shutil.copytree() rather
        # than dir_util.copy_tree() function is called.
        shutil.copytree(src=src_dirname, dst=trg_dirname, **copytree_kwargs)
    # Else if overwriting this target directory with this source directory...
    elif overwrite_policy is DirOverwritePolicy.OVERWRITE:
        # If an iterable of shell-style globs matching ignorable basenames was
        # passed, log a non-fatal warning. Since the dir_util.copy_tree()
        # function fails to support this functionality and we are currently too
        # lazy to do so, a warning is as much as we're willing to give.
        if ignore_basename_globs is not None:
            logs.log_warning(
                'dirs.copy() parameter "ignore_basename_globs" '
                'ignored when parameter "is_overwritable" enabled.')

        # Recursively copy this source to target directory, preserving symbolic
        # links as is. To silently overwrite all conflicting target paths, the
        # dir_util.copy_tree() rather than shutil.copytree() function is called.
        dir_util.copy_tree(
            src_dirname, trg_dirname, preserve_symlinks=1)

    #FIXME: Given how awesomely flexible the manual approach implemented below
    #is, we should probably consider simply rewriting the above two approaches
    #to reuse the exact same logic. It works. It's preferable. Let's reuse it.

    # Else if logging a warning for each target path that already exists, do so
    # by manually implementing recursive directory copying. Sadly, Python
    # provides no means of doing so "out of the box."
    elif overwrite_policy is DirOverwritePolicy.SKIP_WITH_WARNING:
        # Avoid circular import dependencies.
        from betse.util.path import files, paths, pathnames
        from betse.util.type import sequences

        # Passed parameters renamed for disambiguity.
        src_root_dirname = src_dirname
        trg_root_dirname = trg_dirname

        # Basename of the top-level target directory to be copied to.
        trg_root_basename = pathnames.get_basename(src_root_dirname)

        # For the absolute pathname of each recursively visited source
        # directory, an iterable of the basenames of all subdirectories of this
        # directory, and an iterable of the basenames of all files of this
        # directory...
        for src_parent_dirname, subdir_basenames, file_basenames in _walk(
            src_root_dirname):
            # Relative pathname of the currently visited source directory
            # relative to the absolute pathname of this directory.
            parent_dirname_relative = pathnames.relativize(
                src_dirname=src_root_dirname, trg_pathname=src_parent_dirname)

            # If ignoring basenames...
            if ignore_basename_func is not None:
                # Sets of the basenames of all ignorable subdirectories and
                # files of this source directory.
                subdir_basenames_ignored = ignore_basename_func(
                    src_parent_dirname, subdir_basenames)
                file_basenames_ignored = ignore_basename_func(
                    src_parent_dirname, file_basenames)

                # If ignoring one or more subdirectories...
                if subdir_basenames_ignored:
                    # Log the basenames of these subdirectories.
                    logs.log_debug(
                        'Ignoring source "%s/%s" subdirectories: %r',
                        trg_root_basename,
                        parent_dirname_relative,
                        subdir_basenames_ignored)

                    # Remove these subdirectories from the original iterable.
                    # Since the os.walk() function supports in-place changes to
                    # this iterable, this iterable is modified via this less
                    # efficient function rather than efficient alternatives
                    # (e.g., set subtraction).
                    sequences.remove_items(
                        sequence=subdir_basenames,
                        items=subdir_basenames_ignored)

                # If ignoring one or more files...
                if file_basenames_ignored:
                    # Log the basenames of these files.
                    logs.log_debug(
                        'Ignoring source "%s/%s" files: %r',
                        trg_root_basename,
                        parent_dirname_relative,
                        file_basenames_ignored)

                    # Remove these files from the original iterable. Unlike
                    # above, we could technically modify this iterable via
                    # set subtraction: e.g.,
                    #     subdir_basenames -= subdir_basenames_ignored
                    # For orthogonality, we preserve the above approach instead.
                    sequences.remove_items(
                        sequence=file_basenames,
                        items=file_basenames_ignored)

            # Absolute pathname of the corresponding target directory.
            trg_parent_dirname = pathnames.join(
                trg_root_dirname, parent_dirname_relative)

            # Create this target directory if needed.
            make_unless_dir(trg_parent_dirname)

            # For the basename of each non-ignorable file of this source
            # directory...
            for file_basename in file_basenames:
                # Absolute filenames of this source and target file.
                src_filename = pathnames.join(src_parent_dirname, file_basename)
                trg_filename = pathnames.join(trg_parent_dirname, file_basename)

                # If this target file already exists...
                if paths.is_path(trg_filename):
                    # Relative filename of this file. The absolute filename of
                    # this source or target file could be logged instead, but
                    # this relative filename is significantly more terse.
                    filename_relative = pathnames.join(
                        trg_root_basename,
                        parent_dirname_relative,
                        file_basename)

                    # Warn of this file being ignored.
                    logs.log_warning(
                        'Ignoring existing target file: %s', filename_relative)

                    # Ignore this file by continuing to the next.
                    continue

                # Copy this source to target file.
                files.copy(
                    src_filename=src_filename, trg_filename=trg_filename)
    # Else, this overwrite policy is unrecognized. Raise an exception.
    else:
        raise BetseDirException(
            'Overwrite policy "{}" unrecognized.'.format(overwrite_policy))

# ....................{ MAKERS                             }....................
@type_check
def make_unless_dir(*dirnames: str) -> None:
    '''
    Create all passed directories that do *not* already exist, silently ignoring
    those that *do* already exist.

    All nonexistent parents of this directory are also recursively created,
    reproducing the action of the POSIX-compliant ``mkdir -p`` shell command.

    Parameters
    -----------
    pathnames : tuple[str]
        Tuple of the absolute or relative pathnames of all directories to
        create.

    See Also
    -----------
    :func:`make_parent_unless_dir`
        Related function creating the parent directory of this path.
    '''

    # If this directory does *NOT* already exist, create this directory. To
    # support logging, this condition is explicitly tested for. To avoid race
    # conditions (e.g., in the event this directory is created between testing
    # and creating this directory), we preserve the makedirs() keyword argument
    # "exist_ok=True" below.
    for dirname in dirnames:
        if not is_dir(dirname):
            # Log this creation.
            logs.log_debug('Creating directory: %s', dirname)

            # Create this directory if still needed.
            os.makedirs(dirname, exist_ok=True)


@type_check
def make_parent_unless_dir(*pathnames: str) -> None:
    '''
    Create the parent directories of all passed directories that do *not*
    already exist, silently ignoring those that *do* already exist.

    Parameters
    -----------
    pathnames : tuple[str]
        Tuple of the absolute or relative pathnames to create the parent
        directories of.

    See Also
    -----------
    :func:`make_unless_dir`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.path.pathnames import get_dirname, canonicalize

    # Canonicalize each pathname *BEFORE* attempting to get its dirname.
    # Relative pathnames do *NOT* have sane dirnames (e.g., the dirname for a
    # relative pathname "metatron" is the empty string) and hence *MUST* be
    # converted to absolute pathnames first.
    for pathname in pathnames:
        make_unless_dir(get_dirname(canonicalize(pathname)))

# ....................{ MAKERS ~ convenience               }....................
def canonicalize_and_make_unless_dir(dirname: str) -> str:
    '''
    Create the directory with the passed absolute or relative path if this
    directory does not already exist and return the **canonical form** (i.e.,
    unique absolute path) of this path.

    This convenience function chains the lower-level
    :func:`betse.util.path.pathnames.canonicalize` and
    :func:`make_unless_dir` functions.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import pathnames

    # Dirname canonicalized from this possibly non-canonical dirname.
    dirname = pathnames.canonicalize(dirname)

    # Create this directory if needed.
    make_unless_dir(dirname)

    # Return this dirname.
    return dirname


def join_and_make_unless_dir(*partnames: str) -> str:
    '''
    Join (i.e., concatenate) the passed pathnames with the directory separator
    specific to the current platform, create the directory with the resulting
    absolute or relative path, and return this path.

    This convenience function chains the lower-level
    :func:`betse.util.path.pathnames.join` and :func:`make_unless_dir`
    functions.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import pathnames

    # Dirname joined from these pathnames.
    dirname = pathnames.join(*partnames)

    # Create this directory if needed.
    make_unless_dir(dirname)

    # Return this dirname.
    return dirname

# ....................{ ITERATORS                          }....................
@type_check
def iter_basenames(dirname: str) -> SequenceTypes:
    '''
    Sequence of the basenames of all paths (e.g., non-directory files,
    subdirectories) directly contained within the passed directory.

    Parameters
    -----------
    dirname : str
        Absolute or relative path of the parent directory to inspect.

    Returns
    -----------
    SequenceTypes
        Sequence of the basenames of all child paths of this parent directory.
    '''

    # If this directory does *NOT* exist, raise an exception.
    die_unless_dir(dirname)

    # Sequence of all such basenames.
    return os.listdir(dirname)

# ....................{ ITERATORS ~ subdir                 }....................
@type_check
def iter_subdirnames(dirname: str) -> GeneratorType:
    '''
    Generator yielding the pathname of each direct subdirectory of the passed
    parent directory.

    Parameters
    -----------
    dirname : str
        Absolute or relative path of the parent directory to inspect.

    Returns
    -----------
    GeneratorType
        Generator yielding direct subdirectory pathnames.

    Yields
    -----------
    str
        Absolute or relative path of each direct subdirectory of this directory.

    See Also
    -----------
    https://stackoverflow.com/a/25705093/2809027
        StackOverflow answer strongly inspiring this implementation.
    '''

    # Sequence of the basenames of all subdirectories of this directory
    # implemented as follows:
    #
    # * The generator returned by os.walk() is iterated once and then promptly
    #   discarded, yielding:
    #   * The ignorable absolute or relative path of this parent directory,
    #     which we (of course) were passed as input and thus already have.
    #   * This desired sequence.
    #   * An ignorable sequence of the basenames of all files of this directory.
    # * Errors emitted by low-level functions called by os.walk() (e.g.,
    #   os.listdir()) are *NOT* silently ignored.
    _, subdir_basenames, _ = next(_walk(dirname))

    # Return a generator comprehension prefixing each such basename by the
    # absolute or relative path of the parent directory.
    return (
        os_path.join(dirname, subdir_basename)
        for subdir_basename in subdir_basenames
    )


@type_check
def iter_subdir_basenames(dirname: str) -> SequenceTypes:
    '''
    Sequence of the basenames of all direct subdirectories of the passed parent
    directory.

    Parameters
    -----------
    dirname : str
        Absolute or relative path of the parent directory to inspect.

    Returns
    -----------
    SequenceTypes
        Sequence of direct subdirectory basenames of this parent directory.
    '''

    # Sequence of the basenames of all subdirectories of this directory. See the
    # iter_subdirnames() function for further details.
    _, subdir_basenames, _ = next(_walk(dirname))

    # Return this sequence of basenames as is.
    return subdir_basenames

# ....................{ PRIVATE ~ raisers                  }....................
@type_check
def _raise_exception_dir(dirname: str) -> CallableTypes:
    '''
    Closure raising an application-specific exception encapsulating a passed
    OS exception with additional metadata describing the top-level directory
    against which this exception occurred.

    By default, exceptions raised by standard path functions (e.g.,
    :func:`os.walk`) do *not* expose this directory. When functions
    recursively change directories, these exceptions typically have
    non-human-readable messages ala:

        PermissionError: [Errno 13] Permission denied: '__pycache__'

    This closure encapsulates these low-level exceptions with higher-level
    exceptions having human-readable messages ala:

        BetsePathException: [Errno 13] Permission denied: '__pycache__' in
        "/home/leycec/py/betse/betse/util/path/".

    Design
    -----------
    This closure is principally intended to be passed as the value of the
    ``onerror`` parameter accepted by the :func:`os.walk` and :func:`os.fwalk`
    functions, preventing errors emitted by low-level functions called by these
    functions (e.g., :func:`os.listdir`) from being ignored. (By default, these
    functions silently ignore a subset of these errors.)

    Parameters
    -----------
    dirname : str
        Absolute pathname of the top-level directory to raise exceptions against
        in this closure.

    Returns
    -----------
    CallableTypes
        Closure raising such exceptions.
    '''

    @type_check
    def _raise_exception_dir_inner(exception: Exception) -> None:
        '''
        Closure encapsulating an exception passed by low-level path functions
        (e.g., :func:`os.walk`) against the absolute pathname of the top-level
        directory passed to the :func:`_raise_exception_dir` factory function.

        Parameters
        -----------
        exception : Exception
            Low-level exception raised by a low-level path function.

        Raises
        -----------
        BetsePathException
            High-level exception encapsulating this low-level exception.
        '''

        # Encapsulate this low-level exception with a higher-level exception.
        raise BetsePathException(
            '{} in "{}/".'.format(str(exception), dirname)
        ) from exception

    # Return this closure.
    return _raise_exception_dir_inner

# ....................{ PRIVATE ~ walkers                  }....................
# Undocumented os.fwalk() and os.walk() wrappers defaulting to a sane "onerror"
# callable (namely, the _raise_exception_dir() closure defined above), but
# otherwise identical to their standard variants.

def _fwalk(top, *args, **kwargs) -> GeneratorType:

    # If no "onerror" parameter was explicitly passed, default to a sane
    # exception handler.
    if 'onerror' not in kwargs:
        kwargs['onerror'] = _raise_exception_dir(top)

    # Wrap the standard variant of this generator.
    return os.fwalk(top, *args, **kwargs)


def _walk(top, *args, **kwargs) -> GeneratorType:

    # If no "onerror" parameter was explicitly passed, default to a sane
    # exception handler.
    if 'onerror' not in kwargs:
        kwargs['onerror'] = _raise_exception_dir(top)

    # Wrap the standard variant of this generator.
    return os.walk(top, *args, **kwargs)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Test-specific simulation configuration classes wrapping low-level dictionaries
both serialized to and deserialized from on-disk YAML-formatted files.
'''

#FIXME: This submodule requires heavy refactoring away from the current
#low-level approach in favour of the new high-level "confabc"-based approach --
#namely, the "YamlABC" base class coupled with the conf_alias() data
#descriptor. Together, this new functionality provides a substantially better
#YAML-to-Python-object-mapping (YPOM) than the ad-hoc and overly verbose
#boilerplate implemented below. Specifically:
#
#* Eliminate all properties defined below.
#* Replace all usage of the low-level "self._p._conf" dictionary with high-level
#  data descriptors defined by "self._p".
#FIXME: Ideally, after implementing the above, use of the "SimConfigTestWrapper"
#wrapper will be able to be replaced everywhere in tests by direct use of the
#"SimConfTestInternal.p" property providing direct access to this "Parameters"
#object, which increasingly provides all test functionality. We're not quite
#there yet -- but we will be, eventually.

# ....................{ IMPORTS                            }....................
# This subclass necessarily imports from submodules defined by the main codebase
# and is thus *NOT* safely importable in a fixture submodule directly imported
# by a "conftest" plugin module. To defer the importation of this submodule
# until *AFTER* test collection, this submodule is intentionally segregated.
from betse.science.config import confio
from betse.science.config.confenum import IonProfileType, SolverType
from betse.science.phase.require import phasereqs
from betse.science.phase.require.abc.phasereqset import (
    SimPhaseRequirementsOrNoneTypes)
from betse.science.export.csv.csvpipe import SimPipeExportCSVs
from betse.science.visual.anim.animpipe import AnimCellsPipe
from betse.science.visual.plot.pipe.plotpipecell import PlotCellPipe
from betse.science.visual.plot.pipe.plotpipecells import PlotCellsPipe
from betse.util.io.log import logs
from betse.util.type.cls import classes
from betse.util.type.types import type_check

# ....................{ SUPERCLASSES                       }....................
class SimConfigTestWrapper(object):
    '''
    Test-specific simulation configuration wrapper wrapping a low-level
    dictionary deserialized from a YAML-formatted simulation configuration file.

    This wrapper is intended to be instantiated *only* by non-interactive
    automation (e.g., tests, scripts).

    This wrapper superficially wraps this dictionary with convenience methods
    safely modifying this dictionary. This wrapper is lower level than the
    high-level :class:`Parameters` simulation configuration, which transforms
    this dictionary into numerous high-level objects rather than merely wrapping
    this dictionary. While the :class:`Parameters` configuration is principally
    used by backend simulation modelling in the :mod:`betse.science` package,
    this wrapper is principally used by frontend logic modifying simulation
    configurations on behalf of either interactive users (e.g., BETSE's GUI) or
    automated tests.

    Attributes
    ----------
    _p : Parameters
        High-level simulation configuration encapsulated by this test wrapper.
    '''

    # ..................{ MAKERS                             }..................
    #FIXME: Rename to simply make_default().
    @classmethod
    def wrap_new_default(cls, filename: str) -> None:
        '''
        Write the default YAML-formatted simulation configuration to the passed
        path, recursively copy all external resources (e.g., geometry masks)
        referenced and hence required by this configuration into this path's
        directory, and return an instance of this class encapsulating this
        configuration.

        This factory method creates a valid simulation configuration consumable
        by all BETSE CLI commands (e.g., `betse sim`), modified from the default
        simulation configuration shipped with BETSE as follows:

        * The `plot after solving` option in the `results options` section is
          coerced to `False`, preventing hapless end-users from drowning under
          an intimidating deluge of plot windows irrelevant to "beginner" usage.

        Parameters
        ----------
        filename : str
            Absolute or relative path of the simulation configuration file to be
            written. Since this file will be YAML-formatted, this filename
            should ideally be suffixed by a valid YAML filetype: namely, either
            `.yml` or `.yaml`. This is _not_ strictly necessary, but is strongly
            recommended.

        Raises
        ----------
        BetseFileException
            If this file already exists.
        '''

        # Create this YAML file.
        confio.write_default(filename)

        # Create and return an instance of this class wrapping this file.
        return cls(filename)

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, filename: str) -> None:
        '''
        Wrap the low-level dictionary deserialized from the passed
        YAML-formatted simulation configuration file.

        Parameters
        ----------
        filename : str
            Absolute or relative path of this file.
        '''

        # Defer heavyweight imports.
        from betse.science.parameters import Parameters

        # In-memory simulation configuration deserialized from this file.
        self._p = Parameters().load(filename)

    # ..................{ PROPERTIES                         }..................
    # For safety, these properties lack setters and hence are read-only.

    @property
    def p(self) -> 'betse.science.parameters.Parameters':
        '''
        High-level simulation configuration encapsulated by this test wrapper.
        '''

        return self._p

    # ..................{ PROPERTIES ~ path                  }..................
    @property
    def dirname(self) -> str:
        '''
        Absolute or relative path of the directory containing the configuration
        file wrapped by this encapsulation object.
        '''

        return self._p.conf_dirname


    @property
    def filename(self) -> str:
        '''
        Absolute or relative path of the configuration file wrapped by this
        encapsulation object.
        '''

        return self._p.conf_filename

    # ..................{ WRITERS                            }..................
    def overwrite(self) -> None:
        '''
        Silently overwrite the contents of this configuration file with the
        in-memory contents of this configuration dictionary.
        '''

        self._p.save_inplace()

    # ..................{ PROPERTIES ~ float                 }..................
    @property
    def environment_size(self) -> float:
        '''
        Square dimension in meters of the external environment containing the
        cell cluster for this configuration.

        For simplicity, BETSE constrains the environment to be square in shape.
        This dimension thus defines both the environmental width _and_ height.
        '''

        # Coerce the current number to a float for safety.
        return float(self._p._conf['world options']['world size'])


    @environment_size.setter
    @type_check
    def environment_size(self, environment_size: float) -> None:
        '''
        Set the square dimension in meters of the external environment
        containing the cell cluster for this configuration.
        '''

        # Coerce the passed number to a float for safety.
        self._p._conf['world options']['world size'] = float(environment_size)

    # ..................{ MINIMIZERS                         }..................
    def minify(self) -> None:
        '''
        Minimize the space and time costs associated with running the simulation
        configured by this configuration while preserving all fundamental
        configuration features.

        Specifically, this method numerically reduces all configuration options
        pertaining to either world size _or_ simulation time to their **minimum
        permissible values** (i.e., the smallest values still preserving
        simulation stability). This method is intended to be called only by test
        automation.
        '''

        # Minify initialization time to exactly three sampled time steps. For
        # safety, permit currently defined options smaller than the minimums
        # defined below to override these minimums.

        # Duration of each time step in seconds.
        self._p.init_time_step = min(self._p.init_time_step, 1.0e-3)

        # Interval to sample these steps at in seconds.
        self._p.init_time_sampling = min(
            self._p.init_time_sampling, self._p.init_time_step)

        # Total simulation time in seconds. The first digit effectively defines
        # the number of sampled time steps, by the above choice of time step.
        self._p.init_time_total = self._p.init_time_step * 3

        # Minify simulation time to the same durations. To ensure that
        # property setter validation compares durations in the expected manner,
        # properties are assigned in order of increasing duration.
        self._p.sim_time_step = min(
            self._p.sim_time_step, self._p.init_time_step)
        self._p.sim_time_sampling = min(
            self._p.sim_time_sampling, self._p.sim_time_step)
        self._p.sim_time_total = self._p.sim_time_step * 3

        # Minify the physical dimensions of the cell cluster in meters. By
        # experimentation, the default simulation configuration both:
        #
        # * Exhibits expected instabilities for physical dimensions less than
        #   100e-6, typically due to the cell cluster cutting event. Since these
        #   instabilities are expected and hence best ignored, the dimensions
        #   specified below should *NEVER* be smaller than 100e-6.
        # * Exhibits unexpected instabilities for physical dimensions greater
        #   than 250e-6, typically due to non-linear interactions between
        #   channel electroosmosis and physical deformation. Since these
        #   instabilities are unexpected and hence *NOT* safely ignorable, the
        #   dimensions specified below should *ALWAYS* be in this range.
        #
        # Hence, these dimensions reside in an optimal range both avoiding
        # expected instabilities *AND* exposing unexpected instabilities.
        self.environment_size = min(self.environment_size, 100e-6)

        #FIXME: Uncommenting the following line reliably causes the following
        #test to fail with a computational instability during the "sim" phase:
        #
        #    ./test -k test_cli_sim_ecm
        #
        #Since it's unclear whether this behaviour is expected or not, this line
        #remains commented out. (If this behaviour is indeed unexpected, this
        #line should be shifted into the enable_exports_all() method and the
        #more preferable global default retained above).

        # self.environment_size = min(self.environment_size, 250e-6)

        # Minify ECM-specific grid size. For similar reasons as above, the
        # computational grid size specified below appears to be a hard minimum.
        ecm = self._p._conf['general options']
        ecm['comp grid size'] = min(int(ecm['comp grid size']), 20)
        # ecm['plot grid size'] = min(int(ecm['plot grid size']), 50)

        # Log this minification.
        logs.log_debug(
            'Minifying simulation to init '
            'for %f s at [ %f | %f ] s and sim '
            'for %f s at [ %f | %f ] s...',
            self._p.init_time_step,
            self._p.init_time_sampling,
            self._p.init_time_total,
            self._p.sim_time_step,
            self._p.sim_time_sampling,
            self._p.sim_time_total,
        )

    # ..................{ DISABLERS                          }..................
    #FIXME: The implementation of the following methods is fundamentally unsafe.
    #If the structure of the underlying YAML file changes, these methods could
    #silently fail (e.g., if the "plot while solving" option were renamed to
    #"is plotting during"). To combat this, all attempts to directly modify the
    #"self._p._conf" dictionary below *MUST* instead defer to a newly defined
    #set_config_option() method accepting one or more key names followed by the
    #value to set such keys to: e.g.,
    #
    #    set_config_option(('results options', 'plot while solving'), False)
    #
    #If the passed configuration option does *NOT* already exist, that method
    #should raise a human-readable exception. Inevitable future problem solved!

    def disable_visuals(self) -> None:
        '''
        Disable all visual exports, including displaying and saving of all in-
        and post-simulation plots and animations.
        '''

        self._p.anim.is_after_sim = False
        self._p.anim.is_while_sim = False
        self._p.plot.is_after_sim = False


    def disable_interaction(self) -> None:
        '''
        Disable all simulation configuration options either requiring
        interactive user input _or_ displaying graphical output intended for
        interactive user consumption (e.g., plots, animations).

        This method is intended to be called by non-interactive automation
        (e.g., tests, scripts) expecting simulations to behave silently.
        '''

        self._p.anim.is_after_sim_show = False
        self._p.anim.is_while_sim_show = False
        self._p.plot.is_after_sim_show = False

    # ..................{ ENABLERS                           }..................
    def enable_networks(self) -> None:
        '''
        Enable both biochemical reaction and gene regulatory networks.
        '''

        self._p._conf['gene regulatory network settings'][
            'gene regulatory network simulated'] = True

    # ..................{ ENABLERS ~ export                  }..................
    @type_check
    def enable_anim_video(self, writer_name: str, filetype: str) -> None:
        '''
        Enable encoding of all enabled animations as compressed video of the
        passed filetype with the preferred matplotlib animation writer of the
        passed name.

        Parameters
        ----------
        writer_name : str
            Name of the matplotlib animation writer with which to encode video
            (e.g., `ffmpeg`, `imagemagick`).
        filetype : str
            Filetype of videos to encode with this writer (e.g., `mkv`, `mp4`).
        '''

        # Enable animations and animation saving in the general sense.
        self.enable_visuals_save()

        # Localize nested dictionaries for convenience.
        video = self._p._conf['results options']['save']['animations']['video']

        # Enable encoding of the passed filetype with the passed writer type.
        # For determinism, mandate that *ONLY* this writer (rather than two or
        # more writers) be used to do so.
        video['enabled'] = True
        video['filetype'] = filetype
        video['writers'] = [writer_name,]


    def enable_visuals_save(self) -> None:
        '''
        Enable saving of all visual exports, including all in- and
        post-simulation plots and animations.

        This method does *not* enable specific visual exports or simulation
        features required by specific visual exports.
        '''

        self._p.anim.is_while_sim_save = True
        self._p.anim.is_after_sim_save = True
        self._p.plot.is_after_sim_save = True

    # ..................{ ENABLERS ~ solver                  }..................
    def enable_solver_full(self) -> None:
        '''
        Enable the complete BETSE solver.
        '''

        self._p.solver_type = SolverType.FULL


    def enable_solver_circuit(self) -> None:
        '''
        Enable the equivalent circuit-based BETSE solver *and* disable all
        simulation features unsupported by this solver.
        '''

        self._p.solver_type = SolverType.FAST

    # ..................{ ENABLERS ~ solver : circuit        }..................
    def enable_solver_circuit_exports(self) -> None:
        '''
        Enable all possible exports (e.g., CSVs, plots, animations) supported
        by the equivalent circuit-based BETSE solver excluding those requiring
        extracellular spaces, all features required by these exports, and any
        additional features trivially enabled *without* increasing time or space
        complexity.
        '''

        # Enable all simulation features, including the full BETSE solver but
        # excluding extracellular spaces.
        self._enable_solver_circuit_features()

        # Disable extracellular spaces.
        self._p.is_ecm = False

        # Enable all possible exports excluding those requiring the full solver.
        self._enable_exports(requirements_omit=phasereqs.SOLVER_FULL)

    # ..................{ ENABLERS ~ solver : full           }..................
    def enable_solver_full_vg_ions(self) -> None:
        '''
        Enable all voltage-gated ion channels (e.g., sodium, potassium) *and*
        all features required by these channels.

        This method is intended to be called by non-interactive test suites
        exercising these channels. Specifically, this method enables:

        * The full BETSE solver.
        * The extracellular matrix (ECM).
        * The mammalian ion profile (i.e., `animal`), enabling all ions.
        * The intervention increasing the permeability of all cell membranes to
          sodium (Na+).
        * The voltage-gated sodium (Na+) channel `Nav1p2`, corresponding to the
          adult human brain.
        * The voltage-gated potassium (K+) channel `K_Slow`.
        * Decreased time step and sampling rates, ensuring simulation stability.
        * Increased duration and cell count, exposing simulation instabilities.

        For efficiency, this method disables all visuals -- including both in-
        and post-simulation animations and plots.
        '''

        # For efficiency, disable all visuals.
        self.disable_visuals()

        # Enable all features required by these channels.
        self.enable_solver_full()
        self._p.is_ecm = True
        self._p.ion_profile = IonProfileType.MAMMAL

        # For stability, decrease both the time step and sampling rates.
        self._p.sim_time_step     = 1e-4
        self._p.sim_time_sampling = 1e-3

        # For completeness, increase both the duration and cell count.
        self._p.sim_time_total = 50e-3
        self.environment_size = 250e-6

        # Enable the intervention increasing sodium membrane permeability.
        # Although the current default values for this intervention track those
        # defined below fairly closely, the latter are nonetheless explicitly
        # defined below to avoid issues when the former inevitably change.
        sodium_membrane_permeability = self._p._conf['change Na mem']
        sodium_membrane_permeability['event happens'] = True
        sodium_membrane_permeability['change rate']   =  1.0e-3
        sodium_membrane_permeability['change start']  =  5.0e-3
        sodium_membrane_permeability['change finish'] = 30.0e-3
        sodium_membrane_permeability['apply to'] = ['spot',]

        #FIXME: Refactor this to use the new networks formalism.
        # # Enable the voltage-gated sodium (Na+) channel Nav1p2.
        # voltage_gated_sodium_channel = self._p._conf['voltage gated Na+']
        # voltage_gated_sodium_channel['turn on'] = True
        # voltage_gated_sodium_channel['channel type'] = ['Nav1p2',]
        # # voltage_gated_sodium_channel['max value'] = 5.0e-6
        # voltage_gated_sodium_channel['apply to'] = ['base',]
        #
        # # Enable the voltage-gated potassium (K+) channel K_Slow.
        # voltage_gated_potassium_channel = self._p._conf['voltage gated K+']
        # voltage_gated_potassium_channel['turn on'] = True
        # voltage_gated_potassium_channel['channel type'] = ['K_Slow',]
        # # voltage_gated_potassium_channel['max value'] = 5.0e-7
        # voltage_gated_potassium_channel['apply to'] = ['base',]

    # ..................{ ENABLERS ~ solver : full : exports }..................
    def enable_solver_full_exports_ecm(self) -> None:
        '''
        Enable all possible exports (e.g., CSVs, plots, animations) supported
        by the full BETSE solver including those requiring extracellular spaces,
        all features required by these exports, and any additional features
        trivially enabled *without* increasing time or space complexity.
        '''

        # Enable all simulation features, including the full BETSE solver but
        # excluding extracellular spaces.
        self._enable_solver_full_features()

        # Enable extracellular spaces.
        self._p.is_ecm = True

        # Enable all possible exports.
        self._enable_exports()


    def enable_solver_full_exports_noecm(self) -> None:
        '''
        Enable all possible exports (e.g., CSVs, plots, animations) supported
        by the full BETSE solver excluding those requiring extracellular spaces,
        all features required by these exports, and any additional features
        trivially enabled *without* increasing time or space complexity.
        '''

        # Enable all simulation features, including the full BETSE solver but
        # excluding extracellular spaces.
        self._enable_solver_full_features()

        # Disable extracellular spaces.
        self._p.is_ecm = False

        # Enable all possible exports excluding those requiring extracellular
        # spaces.
        self._enable_exports(requirements_omit=phasereqs.ECM)

    # ..................{ PRIVATE ~ enablers                 }..................
    @type_check
    def _enable_exports(
        self,

        # Optional parameters.
        requirements_omit: SimPhaseRequirementsOrNoneTypes = None,
    ) -> None:
        '''
        Enable all possible exports (e.g., CSVs, plots, animations) excluding
        those requiring one or more of the passed simulation phase requirements
        (e.g., the full BETSE solver, extracellular spaces).

        Exports already enabled by the current simulation configuration are
        gracefully preserved as is rather than erroneously readded to their
        respective pipelines.

        Parameters
        ----------
        requirements_omit : SimPhaseRequirementsOrNoneTypes
            Immutable set of all simulation phase requirements such that exports
            requiring one or more requirements in this set are *not* enabled by
            this method. Defaults to ``None``, in which case all possible
            exports are unconditionally enabled.
        '''

        # If unpassed, default the set of requirements to omit to the empty set.
        if requirements_omit is None:
            requirements_omit = phasereqs.NONE

        # For each type of export pipeline to be exercised *AND* a sequence of
        # all exporters enabled by this pipeline...
        for pipe_type, pipe_exporters_enabled in (
            self._pipes_type_exporters_enabled):
            # Name of this pipeline for logging purposes.
            pipeline_name = classes.get_name(pipe_type)

            # Set of the names of all such exporters, for subsequent lookup.
            pipe_exporters_enabled_name = frozenset(
                pipe_exporter_enabled.name
                for pipe_exporter_enabled in pipe_exporters_enabled)
            # logs.log_debug(
            #     'Pipeline "%s" exporters enabled: %r',
            #     pipeline_name, pipe_exporters_enabled_name)

            # For the name of each possible exporter supported by this pipeline,
            # regardless of whether that exporter is currently enabled or not...
            for pipe_exporter_name, pipe_exporter in pipe_type.iter_runners():
                # logs.log_debug(
                #     'Considering pipeline "%s" exporter "%s"...',
                #     pipeline_name, pipe_exporter_name)

                #FIXME: Insufficient. If this exporter is already enabled *BUT*
                #unsatisfied, this exporter *MUST* be manually removed from this
                #pipeline. Trivial, but we need to actually do it.

                # If this exporter is already enabled by this pipeline...
                if pipe_exporter_name in pipe_exporters_enabled_name:
                    # Log this continuation.
                    logs.log_debug(
                        'Ignoring pipeline "%s" exporter "%s", '
                        'as already enabled...',
                        pipeline_name, pipe_exporter_name)

                    # Continue to the next possible exporter.
                    continue
                # Else if this exporter requires one or more simulation features
                # explicitly excluded by the caller...
                elif pipe_exporter.requirements.isintersection(
                    requirements_omit):
                    # Log this exclusion.
                    logs.log_debug(
                        'Excluding pipeline "%s" exporter "%s", '
                        'as test requirements unsatsified...',
                        pipeline_name, pipe_exporter_name)

                    # Continue to the next possible exporter.
                    continue
                # Else, this exporter is *NOT* already enabled by this pipeline
                # and does *NOT* require omitted simulation features.

                # Log this inclusion.
                logs.log_debug(
                    'Including pipeline "%s" exporter "%s"...',
                    pipeline_name, pipe_exporter_name)

                # New default export of this type appended to this pipeline.
                pipe_exporter_conf = pipe_exporters_enabled.append_default()
                pipe_exporter_conf.name = pipe_exporter_name

    # ..................{ PRIVATE ~ enablers : solver        }..................
    def _enable_solver_circuit_features(self) -> None:
        '''
        Enable all simulation features required by all exports (e.g., CSVs,
        plots, animations) supported by the equivalent circuit-based BETSE
        solver, excluding extracellular spaces.

        This method additionally enables optional settings improving test
        coverage but *not* explicitly required by these exports. Specifically,
        this method enables:

        * The full BETSE solver.
        * Saving of all visual exports.
        '''

        # Enable the equivalent circuit-based solver.
        self.enable_solver_circuit()

        # Enable saving of these exports.
        self.enable_visuals_save()


    def _enable_solver_full_features(self) -> None:
        '''
        Enable all simulation features required by all exports (e.g., CSVs,
        plots, animations) -- including the full BETSE solver but excluding
        extracellular spaces.

        This method additionally enables optional settings improving test
        coverage but *not* explicitly required by these exports. Specifically,
        this method enables:

        * The full BETSE solver.
        * Saving of all visual exports.
        * Cell enumeration, labelling each cell by its 0-based index.
        * Current overlays, displaying current density streamlines.
        * All ion concentration plots and animations (e.g., calcium (Ca+),
          hydrogen (H+; pH)) by enabling:
          * The mammalian ion profile (i.e., ``animal``), enabling all ions.
        * The deformation plot and animation by enabling:
          * Galvanotaxis (i.e., deformations).
        * The ``fluid_intra`` plot and animation by enabling:
          * Fluid flow.
        * The ``pressure_mechanical`` plot and animation by enabling:
          * The mechanical pressure event.
        * The ``pressure_osmotic`` plot and animation by enabling:
          * Osmotic pressure.
        * The ``pump_density`` plot and animation by enabling:
          * Channel electroosmosis.
        * The ``voltage_polarity`` plot and animation by enabling:
          * Cell polarizability.
        * All other plots and animations *not* requiring extracellular spaces.
        '''

        # Enable the full solver.
        self.enable_solver_full()

        # Enable saving of these exports.
        self.enable_visuals_save()

        # Localize nested dictionaries for convenience.
        results = self._p._conf['results options']
        variable = self._p._conf['variable settings']

        # Enable all simulation features required by these exports.
        self._p.ion_profile = IonProfileType.MAMMAL
        self._p._conf['apply pressure']['event happens'] = True
        # variable['channel electroosmosis']['turn on'] = True  # This feature has been removed
        variable['deformation']['turn on'] = True
        variable['fluid flow']['include fluid flow'] = True
        variable['pressures']['include osmotic pressure'] = True
        self._p.cell_polarizability = 1e-4

        # Enable all optional settings supported by these exports.
        results['enumerate cells'] = True
        results['overlay currents'] = True

    # ..................{ PRIVOTE ~ iterators                }..................
    #FIXME: Add the CSV export pipeline, once completed, to this tuple.
    @property
    def _pipes_type_exporters_enabled(self) -> tuple:
        '''
        Tuple of 2-tuples ``(pipe_type, pipe_exporters_enabled)``, describing
        each export pipeline to be exercised, where:

        * ``pipe_type`` is the subclass of the :class:`SimPipeABC` superclass
          implementing this pipeline in the codebase.
        * ``pipe_exporters_enabled`` is an instance of the :class:`YamlList`
          class listing all currently enabled exporters in this pipeline.
        '''

        return (
            (SimPipeExportCSVs, self._p.csv.csvs_after_sim),
            (PlotCellPipe,  self._p.plot.plots_cell_after_sim),
            (PlotCellsPipe, self._p.plot.plots_cells_after_sim),
            (AnimCellsPipe, self._p.anim.anims_after_sim),
        )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2017 by Alexis Pietak & Cecil Curry
# See "LICENSE" for further details.

import os.path
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.collections import LineCollection, PolyCollection
from betse.exceptions import BetseExceptionSimulation
from betse.science import filehandling as fh
from betse.science.cells import Cells
from betse.science.parameters import Parameters
from betse.science.plot import plot as viz
from betse.science.sim import Simulator
from betse.science.tissue.handler import TissueHandler
from betse.util.io.log import logs
from betse.util.path import files, paths
from betse import ignition, pathtree

# Initialize the current application before doing anything else.
ignition.init()
# at the moment only the default config file can be used
config_filename = pathtree.CONFIG_DEFAULT_FILENAME

# Validate and localize such filename.
files.die_unless_file(config_filename)
_config_filename = config_filename
_config_basename = paths.get_basename(_config_filename)

logs.log_info(
    'Plotting cell cluster with configuration file "{}".'.format(_config_basename))

p = Parameters(config_filename=_config_filename)  # create an instance of Parameters
p.I_overlay = False  # force the current overlay to be false as there's no data for it
sim = Simulator(p)

cells = Cells(p, worldtype='basic')

if files.is_file(cells.savedWorld):
    cells, _ = fh.loadWorld(cells.savedWorld)  # load the simulation from cache
    p.sim_ECM = cells.sim_ECM
    logs.log_info('Cell cluster loaded.')
else:
    raise BetseExceptionSimulation("Ooops! No such cell cluster file found to load!")

if p.sim_ECM is False:
    sim.baseInit_all(cells, p)
    dyna = TissueHandler(sim, cells, p)
    dyna.tissueProfiles(sim, cells, p)
else:
    sim.baseInit_all(cells, p)
    dyna = TissueHandler(sim, cells, p)
    dyna.tissueProfiles(sim, cells, p)

if p.autosave is True:
    images_path = p.init_results
    image_cache_dir = os.path.expanduser(images_path)
    os.makedirs(image_cache_dir, exist_ok=True)
    savedImg = os.path.join(image_cache_dir, 'fig_')

fig_tiss, ax_tiss, cb_tiss = viz.clusterPlot(
    p, dyna, cells, clrmap=p.default_cm)

if p.autosave is True:
    savename10 = savedImg + 'cluster_mosaic' + '.png'
    plt.savefig(savename10, format='png', transparent=True)

if p.turn_all_plots_off is False:
    plt.show(block=False)

if p.sim_ECM is True:
    plt.figure()
    ax99 = plt.subplot(111)
    plt.imshow(
        np.log10(sim.D_env_weight.reshape(cells.X.shape)),
        origin='lower',
        extent=[p.um * cells.xmin, p.um * cells.xmax, p.um * cells.ymin, p.um * cells.ymax],
        cmap=p.default_cm,
    )
    plt.colorbar()

    cell_edges_flat = p.um * cells.mem_edges_flat
    coll = LineCollection(cell_edges_flat, colors='k')
    coll.set_alpha(1.0)
    ax99.add_collection(coll)

    plt.title('Logarithm of Environmental Diffusion Weight Matrix')

    if p.autosave is True:
        savename10 = savedImg + 'env_diffusion_weights' + '.png'
        plt.savefig(savename10, format='png', transparent=True)

    if p.turn_all_plots_off is False:
        plt.show(block=False)

    plt.figure()
    plt.imshow(
        cells.maskM,
        origin='lower',
        extent=[p.um * cells.xmin, p.um * cells.xmax, p.um * cells.ymin, p.um * cells.ymax],
        cmap=p.default_cm,
    )
    plt.colorbar()
    plt.title('Cluster Masking Matrix')

    if p.autosave is True:
        savename = savedImg + 'cluster_mask' + '.png'
        plt.savefig(savename, format='png', transparent=True)

    if p.turn_all_plots_off is False:
        plt.show(block=False)

# plot gj
fig_x = plt.figure()
ax_x = plt.subplot(111)

if p.showCells is True:
    base_points = np.multiply(cells.cell_verts, p.um)
    col_cells = PolyCollection(base_points, facecolors='k', edgecolors='none')
    col_cells.set_alpha(0.3)
    ax_x.add_collection(col_cells)

con_segs = cells.nn_edges
connects = p.um * np.asarray(con_segs)
collection = LineCollection(connects, linewidths=1.0, color='b')
ax_x.add_collection(collection)
plt.axis('equal')
plt.axis([cells.xmin * p.um, cells.xmax * p.um, cells.ymin * p.um, cells.ymax * p.um])

ax_x.set_xlabel('Spatial x [um]')
ax_x.set_ylabel('Spatial y [um')
ax_x.set_title('Cell Connectivity Network')

if p.autosave is True:
    savename10 = savedImg + 'gj_connectivity_network' + '.png'
    plt.savefig(savename10, format='png', transparent=True)

if p.turn_all_plots_off is False:
    plt.show(block=False)
    plt.show()

else:
    logs.log_info(
        'Plots exported to init results folder '
        'defined in configuration file "{}".'.format(_config_basename))


#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Global test configuration for all tests.

:mod:`pytest` implicitly imports *all* functionality defined by this module into
*all* test modules.
'''

# ....................{ IMPORTS                            }....................
import pytest
from betse import metadata
from betse.util.os.shell import shellenv
from betse_test.exceptions import BetseTestHookException
from betse_test.util.testabc import SerialTestABC

# ....................{ IMPORTS ~ fixture                  }....................
from betse_test.fixture.igniter import betse_init
from betse_test.fixture.tempdirer import betse_temp_dir
from betse_test.fixture.simconf.simconfer import (
    betse_sim_conf,
    betse_sim_conf_default,
    betse_sim_conf_compat,
)

# ....................{ HOOKS ~ plugin                     }....................
def pytest_configure(config):
    '''
    Hook run immediately *after* both parsing all :mod:`pytest` command-line
    options and loading all third-party :mod:`pytest` plugins (including
    application-specific ``conftest`` scripts).

    Specifically:

    * The BETSE-specific :attr:`betse._is_pytest` global boolean is set to
      ``True``, informing the main codebase that tests are currently being run.
      Logic elsewhere then performs test-specific handling if this boolean is
      enabled (e.g., defaulting to a non-interactive matplotlib backend suitable
      for usage in this possibly non-interactive test environment).
    * If the external ``${DISPLAY}`` environment variable is currently set
      (e.g., to the X11-specific socket to be connected to display GUI
      components), unset this variable. Permitting this variable to remain set
      would permit tests erroneously attempting to connect to an X11 server to
      locally succeed but remotely fail, as headless continuous integration (CI)
      typically has no access to an X11 server. Unsetting this variable ensures
      orthogonality between these cases by coercing the former to fail as well.
    '''

    #FIXME: The following two operations should be converted into autouse
    #fixtures defined in this plugin above. Such fixtures should require the
    #builtin fixture permitting us to temporarily change environment variables,
    #which should then be used to temporarily undefine the ${DISPLAY} variable.
    #What was that called again... "monkeypatch"? Contemplate eternity and see:
    #
    #    http://pytest.org/latest/fixture.html#autouse-fixtures-xunit-setup-on-steroids

    # Inform the main codebase that tests are currently being run.
    metadata._IS_TESTING = True

    # Unset the external `${DISPLAY}` environment variable if currently set.
    # Technically, this operation needs to be performed:
    #
    # * Only once for the entire test suite when py.test is *NOT* parallelized
    #   with "xdist", in which case all tests run in the same process and hence
    #   share the same environment variables.
    # * Once for each test when py.test is parallelized with "xdist", in which
    #   case each test is run in a distinct subprocess and hence does *NOT*
    #   share the same environment variables.
    #
    # Since unsetting environment variables is fast, doing so here transparently
    # supports both use cases detailed above with no discernable downside. See
    # the docstring for additional commentary.
    shellenv.unset_var_if_set('DISPLAY')


def pytest_unconfigure(config):
    '''
    Hook run immediately *before* exiting the current :mod:`pytest` test
    session.

    Specifically:

    * The BETSE-specific :attr:`betse._is_pytest` global boolean is set to
      ``False``, informing the main codebase that tests are no longer currently
      being run.
    '''

    metadata._IS_TESTING = False

# ....................{ HOOKS ~ test                       }....................
def pytest_runtest_setup(item: 'pytest.main.Item') -> None:
    '''
    Hook run immediately *before* running the passed test.

    Specifically:

    * If this is a **serial test** (i.e., test method bound to an instance of
      the :class:`SerialTestABC` superclass) for which a prior serial test in
      the same test class was recorded as failing by the
      :func:`pytest_runtest_makereport` hook, this hook marks this test as
      xfailing.

    Parameters
    ----------
    item : pytest.main.Item
        Metadata encapsulating this test callable (e.g., function, method).

    See Also
    ----------
    https://pytest.org/latest/example/simple.html#incremental-testing-test-steps
        Official py.test code snippet inspiring this implementation.
    '''

    # For each list of fixtures requested by this test...
    for fixture_defs in item._fixtureinfo.name2fixturedefs.values():
        # For each such fixture...
        for fixture_def in fixture_defs:
            # True only if this fixture is parametrized serially (i.e., is
            # decorated by the @serialize_parametrized_fixture decorator).
            is_fixture_parametrized_serially = getattr(
                fixture_def.func,
                '_betse_is_fixture_parametrized_serially',
                False)
            # print('funcinfo: {}'.format(fixture_def))
            # print('funcinfo.func: {}'.format(fixture_def.func))
            # print('funcinfo.func.is: {}'.format(is_fixture_parametrized_serially))

            # If this fixture is parametrized serially, propagate this setting
            # to this test by dynamically marking this test accordingly.
            if is_fixture_parametrized_serially:
                item.keywords['serialize_parametrized_test'] = True

    # If this test is parametrized serially, do so *AFTER* propagating this
    # setting from fixtures requested by this test above.
    if 'serialize_parametrized_test' in item.keywords:
        # If this test is *NOT* parametrized, raise an exception.
        if not hasattr(item, '_genid'):
            raise BetseTestHookException(
                'Unparametrized test "{}" marked as serial.'.format(item.name))

        # Test callable (i.e., underlying function or method object).
        test_callable = item.obj

        # Unique identifier of the first failing parameters passed to this test
        # if any or "None" otherwise.
        first_failing_param_id = getattr(
            test_callable, '_betse_first_failing_param_id', None)

        # If any parameters passed to this test failed, mark these subsequently
        # passed parameters as also xfailing.
        if first_failing_param_id is not None:
            pytest.xfail(
                'Prior serial test parametrization "{}" failed.'.format(
                    first_failing_param_id))
    #FIXME: This should *REALLY* simply be implemented as a new @serialize_class
    #decorator, much like @serialize_parametrized_test above.
    # If this is an unparametrized test intended to be run serially, do so.
    elif SerialTestABC.is_test_serial(item):
        # Object to which this test method is bound.
        test_instance = item.parent

        # Name of the first previously declared test method to fail in this test
        # method's class if any or "None" otherwise.
        first_failure_method_name = getattr(
            test_instance, '_first_failure_method_name', None)

        # If a previously declared test failed, mark this subsequently declared
        # test as also xfailing.
        if first_failure_method_name is not None:
            pytest.xfail('Prior serial test "{}" failed.'.format(
                test_instance._first_failure_method_name))


def pytest_runtest_makereport(
    item: 'pytest.main.Item', call: 'pytest.runner.CallInfo'):
    '''
    Hook run immediately *after* the passed test returned the passed result.

    If this is a **serial test** (i.e., test method bound to an instance of the
    the :class:`SerialTestABC` superclass) that failed, this hook records this
    failure for subsequent analysis by the :func:`pytest_runtest_setup` hook.

    Parameters
    ----------
    item : pytest.main.Item
        Metadata encapsulating this test callable (e.g., function, method).
    call : pytest.runner.CallInfo
        Metadata encapsulating the value returned or exception raised by calling
        this test callable.

    See Also
    ----------
    https://pytest.org/latest/example/simple.html#incremental-testing-test-steps
        Official py.test code snippet inspiring this implementation.
    '''

    # If this test failed...
    if call.excinfo is not None:
        # If this is a parametrized test intended to be run serially...
        if 'serialize_parametrized_test' in item.keywords:
            # Test callable (i.e., underlying function or method object).
            test_callable = item.obj

            # Record the unique identifiers of these test parameters for
            # subsequent use by the pytest_runtest_setup() hook.
            test_callable._betse_first_failing_param_id = item._genid
        #FIXME: This should *REALLY* simply be implemented as a new
        #@class_serial decorator, much like @parametrize_serial above.
        # If this is an unparametrized test intended to be run serially...
        elif SerialTestABC.is_test_serial(item):
            # Object to which this test method is bound.
            test_parent = item.parent

            # Record the name of this test for subsequent use by the
            # pytest_runtest_setup() hook.
            test_parent._first_failure_method_name = item.name



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Application-specific ``test`` subcommand for :mod:`setuptools`.
'''

# ....................{ IMPORTS                            }....................
from betse.util.type.obj import objects
from betse_setup import buputil
from distutils.errors import DistutilsClassError
from setuptools import Command

# ....................{ COMMANDS                           }....................
def add_setup_commands(metadata: dict, setup_options: dict) -> None:
    '''
    Add the ``test`` subcommand to the passed dictionary of :mod:`setuptools`
    options.
    '''

    buputil.add_setup_command_classes(metadata, setup_options, test)

# ....................{ CLASSES ~ base                     }....................
class test(Command):
    '''
    Command class testing the current application with :mod:`pytest`.

    Attributes
    ----------
    match_name : str
        Python-evaluatable expression (e.g., ``'test_method or test other'``)
        conditionally matching a substring of the name of each test function or
        class to be run if any *or* `None` if all tests are to be run
        unconditionally.
    _pytest_public : PackageType
        Top-level public :mod:`pytest` package, imported by the :meth:`run`
        method.
    _pytest_private : PackageType
        Top-level private :mod:`_pytest` package, imported by the :meth:`run`
        method.

    See Also
    ----------
    https://github.com/pytest-dev/pytest-runner
        Recommended implementation of a :mod:`setuptools` command invoking
        :mod:`pytest`, which is overly obsessed with egg-based local
        installation of test dependencies rather than actual testing.
        Naturally, this is less than useful.
    https://pytest.org/latest/goodpractices.html#manual-integration
        Fallback implementation for a :mod:`setuptools` command invoking
        :mod:`pytest`, which is sufficiently trivial as to be effectively
        useless. This is also less than useful.
    '''

    # ..................{ ATTRIBUTES                         }..................
    description = 'run py.test-driven functional and unit tests'
    '''
    Command description printed on running ``./setup.py --help-commands``.
    '''


    user_options = [
        # Boolean options.
        (
            'no-capture', 's',
            'Prevent py.test from silently capturing any '
            'standard error, standard output, or logging messages '
            'emitted by tests. '
            '(By default, py.test silently captures all three.) '
            'All three will be written in real-time "as is" to their '
            'respective file handles (e.g., current terminal, logfile).'
        ),

        # Argument options.
        (
            'match-name=', 'k',
            'Only run tests which match the given substring expression. '
            'An expression is a python evaluatable expression '
            'where all names are substring-matched against '
            'test names and their parent classes. '
            'Example: -k "test_method or test other" matches '
            'all test functions and classes whose names contain '
            'either "test_method" or "test_other". '
            'Keywords are also matched to classes and functions containing '
            'extra names in their "extra_keyword_matches" set as well as to '
            'functions which have names assigned directly to them.'
        ),
        (
            'export-sim-conf-dir=', None,
            'Target directory into which all '
            'source simulation configuration directories produced by '
            '"@skip_unless_export_sim_conf"-marked tests are to be copied.'
        ),
    ]
    '''
    List specifying command-line options accepted by this command.

    Each item of this list is a 3-tuple ``(long_name, short_name, synopsis)``
    specifying a single option, where:

    * ``long_name`` is this option's mandatory GNU-style long form excluding
      ``--`` prefix. Additionally:
      * If this option accepts a mandatory argument, this long form should be
        suffixed by ``=``.
      * Else, this long form should *not* be suffixed by ``=``.
    * ``short_name`` is either:
      * This option's option *NIX-style short form excluding ``-`` prefix.
      * ``None`` if this option has no such form.
    * ``synopsis`` is this option's mandatory help documentation.

    For each such option, an attribute of the same name as this option's long
    form *must* be explicitly initialized in the :meth:`initialize_options`
    method. :mod:`setuptools` fails to recognize options for which this is
    *not* the case. (You fail a simple sanity check yet again,
    :mod:`setuptools`.)

    See Also
    ----------
    http://ilostmynotes.blogspot.ca/2009/04/python-distutils-installer-and.html
        Inarguably, the best (albeit unofficial) documentation on this list.
    '''

    # ..................{ SUPERCLASS                         }..................
    def initialize_options(self):
        '''
        Declare option-specific attributes subsequently initialized by
        :meth:`finalize_options`.

        If this function is *not* defined, the default implementation of this
        method raises an inscrutable :mod:`distutils` exception. If these
        attributes are *not* declared, the subsequent call to
        :meth:`self.set_undefined_options` raises an inscrutable
        :mod:`setuptools` exception. (This is terrible. So much hate.)
        '''

        # Option-specific public attributes. For each option declared by the
        # "user_options" list above, a public attribute of the same name as this
        # option's long form *MUST* be initialized here to its default value.
        self.no_capture = None
        self.match_name = None
        self.export_sim_conf_dir = None

        # setuptools-specific public attributes.
        # self.install_dir = None

        # Custom private attributes.
        self._pytest_public = None
        self._pytest_private = None


    def finalize_options(self):
        '''
        Default undefined command-specific options to the options passed to the
        current parent command if any (e.g., ``symlink``).
        '''

        pass


    def run(self):
        '''
        Run the current command and all subcommands thereof.
        '''

        # Import and monkey-patch "py.test" *BEFORE* running "py.test".
        self._init_pytest()
        self._patch_pytest()
        self._run_pytest()

    # ..................{ PRIVATE                            }..................
    def _init_pytest(self) -> None:
        '''
        Dynamically import the top-level :mod:`pytest` package into this
        object's :attr:`_pytest` instance variable.

        Specifically, this method imports the top-level `_pytest.main` module,
        providing programmatic access to py.test's CLI implementation. While
        py.test is also runnable as an external command, doing so invites
        non-trivial complications. Unlike most Python applications (e.g.,
        PyInstaller), py.test is *not* dynamically importable via the following
        import machinery:

            # Don't do this.
            >>> pytest_main = buputil.import_module(
            ...    'pytest.main', exception_message=(
            ...    'py.test not installed under the current Python interpreter.'))

        Attempting to do so reliably raises exceptions resembling:

            AttributeError: 'module' object has no attribute '__path__'

        The reason why appears to be package ``__path__`` manipulations
        dynamically performed by the private :mod:`_pytest` package. Sadly,
        attempting to import :func:`_pytest.main` fails with an even more
        inscrutable exception. The solution, of course, is to brute-force it
        by only dynamically importing the root :mod:`pytest` package.
        '''

        # Import the public "pytest" package *BEFORE* the private "_pytest"
        # package. While importation order is typically ignorable, imports can
        # technically have side effects. Tragicomically, this is the case here.
        # Importing the public "pytest" package establishes runtime
        # configuration required by submodules of the private "_pytest" package.
        # The former *MUST* always be imported before the latter. Failing to do
        # so raises obtuse exceptions at runtime... which is bad.
        self._pytest_public = buputil.import_module(
            'pytest', exception_message=(
            'py.test not installed under the current Python interpreter.'))
        self._pytest_private = buputil.import_module(
            '_pytest', exception_message=(
            'py.test not installed under the current Python interpreter.'))


    def _patch_pytest(self) -> None:
        '''
        Monkey-patch the :mod:`pytest` framework in the active Python
        interpreter, altering startup :mod:`pytest` functionality in an
        early-time manner *not* permitted within :mod:`pytest` plugins.

        `pytest` plugins (e.g., ``conftest`` submodules of a test suite) are
        imported by :mod:`pytest` *after* :mod:`pytest` startup and hence cannot
        be used to alter startup :mod:`pytest` functionality in an early-time
        manner.

        Specifically, this method monkey-patches:

        * The :meth:`CaptureManager._getcapture` method to capture stderr but
          *not* stdout (rather than neither stderr nor stdout) when the
          ``py.test`` command is passed either the ``-s`` or ``--capture=no``
          CLI options. The default approach of *not* capturing stderr prevents
          ``py.test`` from capturing and hence reporting error messages in
          failure reports, requiring tedious upwards scrolling through test
          output to find the corresponding error messages.
        '''

        # py.test classes to be monkey-patched.
        from _pytest.capture import CaptureManager, FDCapture, MultiCapture

        # If the private method to be monkey-patched no longer exists, py.test
        # is either broken or unsupported. In either case, raise an exception.
        if not objects.is_method(CaptureManager, '_getcapture'):
            raise DistutilsClassError(
                'Method pytest.capture.CaptureManager._getcapture() '
                'not found. The current version of py.test is either '
                'broken (unlikely) or unsupported (likely).'
            )

        # Old method to be monkey-patched.
        _getcapture_old = CaptureManager._getcapture

        # New method applying this monkey-patch.
        def _getcapture_new(self, method):
            if method == "no":
                return MultiCapture(
                    out=False, err=True, in_=False, Capture=FDCapture)
            else:
                return _getcapture_old(self, method)

        # Replace the old with the new method.
        CaptureManager._getcapture = _getcapture_new


    def _run_pytest(self) -> None:
        '''
        Call the :func:`pytest.main` function in the active Python interpreter,
        passed CLI options corresponding to the CLI options passed to this
        setuptools command.
        '''

        # List of all shell words to be passed as arguments to "py.test".
        #
        # Whereas the top-level "pytest.ini" configuration file lists options
        # unconditionally passed to *ALL* "py.test" invocations, this list only
        # lists options conditionally applicable to invocations run within the
        # "python3 setup.py test" subcommand.
        #
        # For example, the "--capture=no" option should be conditionally enabled
        # *ONLY* when applying the monkey-patch applied by this setuptools
        # subcommand (namely, here). This option is intentionally omitted from
        # the top-level "pytest.ini" configuration file.
        pytest_args = [
            # When testing interactively, prevent py.test from capturing stdout
            # but *NOT* stderr. By default, py.test captures and delays printing
            # stdout until after test completion. While a possibly suitable
            # default for short-lived unit tests, such capturing is unsuitable
            # for long-lived functional tests.
            #
            # Note that this option is monkey-patched by the _patch_pytest()
            # method to capture only stdout. By default, this option captures
            # neither stdout (which is good) nor stderr (which is bad).
            '--capture=no',

            # When testing interactively, halt testing on the first failure.
            # Permitting multiple failures complicates failure output,
            # especially when every failure after the first is a result of the
            # same underlying issue.
            #
            # When testing non-interactively, testing is typically *NOT* halted
            # on the first failure. Hence, this option is confined to this
            # subcommand rather than added to the general-purpose "pytest.ini"
            # configuration.
            '--maxfail=1',
        ]

        #FIXME: Disable "xdist" if at least one serial test exists. Currently,
        #none do. Theoretically, they could. They once did and certaily could
        #again. Serial functional tests assume that all tests to be serialized
        #are run in the same Python process. "pytest-xdist" currently provides
        #no means of doing so; instead, "pytest-xdist" assigns all tests to
        #arbitrary test slaves and hence Python processes. Until "pytest-xdist"
        #permits tests to be isolated to the same test slave, "pytest-xdist"
        #must *NOT* be enabled. Non-fatal warning to output might resemble:
        #    buputil.output_warning(
        #        'py.test plugin "pytest-xdist" fails to support serial tests.')
        #    buputil.output_warning(
        #        'Tests will *NOT* be parallelized across multiple processors.')

        # True only if the optional third-party "pytest-xdist" plugin is both
        # importable and *NOT* explicitly disabled below (e.g., due to the end
        # user having passed CLI options incompatible with this plugin).
        is_xdist = buputil.is_module('xdist')

        #FIXME: Disabled for the moment. "xdist" appears to be inexplicably
        #failing with non-human-readable exceptions. The lack of official
        #support for test parallelization in py.test is becoming clear.
        #
        #When "xdist" is (eventually) stabilized, excise this reassignment here
        #*AND* reenable the conditional below.
        is_xdist = False

        #FIXME: Reenable this conditional.
        # If this plugin is unimportable, print a non-fatal warning. Due to the
        # cost of running tests, parallelization is highly recommended.
        # if not is_xdist:
        #     buputil.output_warning(
        #         'Optional py.test plugin "pytest-xdist" not found.')
        #     buputil.output_warning(
        #         'Tests will *NOT* be parallelized across multiple processors.')

        # Pass options passed to this subcommand to this py.test command,
        # converting long option names specific to this subcommand (e.g.,
        # "--no-capture") to short option names recognized by py.test (e.g.,
        # "-s"). Sadly, py.test typically recognizes only the latter.
        #
        # If the "-s" option is passed...
        if self.no_capture is not None:
            # If "pytest-xdist" is importable, print a non-fatal warning.
            # This plugin silently ignores all capture redirection CLI options,
            # including "-s", "--no-capture", and "--capture" options. It
            # appears unlikely that this will ever be solved. For details, see:
            # https://github.com/pytest-dev/pytest/issues/680
            if is_xdist:
                # Print a non-fatal warning.
                buputil.output_warning(
                    'Option "-s" unsupported by py.test plugin "pytest-xdist".')
                buputil.output_warning(
                    'Tests will *NOT* be parallelized across '
                    'multiple processors.')

                # Disable this plugin. While parallelization is important,
                # respecting end user wishes with respect to capture redirection
                # is paramount and hence takes precedence.
                is_xdist = False

            # Pass the "-s" option to py.test.
            pytest_args.append('-s')
            # pytest_args.append('--capture=no')

        # If any remaining option is passed, forward these option unmodified
        # onto "py.test". For safety, these option's arguments are intentionally
        # *NOT* shell-quoted. Doing so unnecessarily adds an additional level of
        # quoting... which is bad.
        if self.match_name is not None:
            pytest_args.extend(('-k', self.match_name,))
        if self.export_sim_conf_dir is not None:
            pytest_args.extend(
                ('--export-sim-conf-dir', self.export_sim_conf_dir,))

        # If "pytest-xdist" is both importable and *NOT* explicitly disabled...
        if is_xdist:
            # Notify the user of test parallelization.
            print('Optional py.test plugin "pytest-xdist" found.')
            print('Tests will be parallelized across all available processors.')

            # Instruct "pytest-xdist" to autodetect and parallelize tests to all
            # available processors.
            pytest_args.extend(['-n', 'auto'])

        # Instruct "py.test" of the relative pathname to the top-level directory
        # for this project. On startup, "py.test" internally:
        #
        # * Sets its "rootdir" property to this pathname in absolute form.
        # * Sets its "inifile" property to the concatenation of this pathname
        #   with the basename "pytest.ini" if this top-level configuration file
        #   exists.
        # * Prints the initial values of these properties to stdout.
        #
        # *THIS IS ESSENTIAL.* If "py.test" is *NOT* explicitly passed this
        # relative pathname as an argument, "py.test" typically fails to set
        # these properties to the expected paths. For unknown reasons
        # (presumably unresolved "py.test" issues), "py.test" instead sets
        # "rootdir" to the absolute path of the current user's home directory
        # and "inifile" to "None" -- because no current user's home directory
        # contains a "pytest.ini" file. The error output resembles:
        #
        #    $ ./test -k test_sim_export --export-sim-conf-dir ~/tmp/yolo
        #    running test
        #    Running py.test with arguments: ['--capture=no', '--maxfail=1', '-k', 'test_sim_export', '--export-sim-conf-dir', '/home/leycec/tmp/yolo']
        #    usage: setup.py [options] [file_or_dir] [file_or_dir] [...]
        #    setup.py: error: unrecognized arguments: --export-sim-conf-dir
        #      inifile: None
        #      rootdir: /home/leycec
        #
        # See the following official documentation for further details, entitled
        # "Initialization: determining rootdir and inifile":
        #     https://docs.pytest.org/en/latest/customize.html
        pytest_args.append('.')

        # Fork the "py.test" command, propagating its exit status as our own.
        print('Running py.test with arguments: {}'.format(pytest_args))
        buputil.exit_with_status(self._pytest_public.main(pytest_args))



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **non-string iterable** (i.e., non-string object implementing the
abstract base class `collections.abc.Iterable`) facilities.

See Also
----------
:func:`betse.util.type.types.is_iterable`
    Further details on what constitutes iterables and non-string iterables.
'''

# ....................{ IMPORTS                           }....................
import itertools
from betse.exceptions import BetseIterableException
from betse.util.type import types
from betse.util.type.types import (
    type_check,
    CallableTypes,
    ClassType,
    GeneratorType,
    IterableTypes,
    MappingType,
    SizedType,
    TestableTypes,
)
from collections import deque
from operator import itemgetter

# ....................{ EXCEPTIONS                        }....................
@type_check
def die_unless_items_instance_of(
    iterable: IterableTypes, cls: TestableTypes) -> None:
    '''
    Raise an exception unless *all* items of the passed iterable are instances
    of the passed class or tuple of classes.

    Parameters
    ----------
    iterable: IterableTypes
        Iterable to be validated.
    cls : TestableTypes
        Class or tuple of classes to validate that all items of this iterable
        be instances of.

    Raises
    ----------
    BetseIterableException
        If at least one item of this iterable is *not* an instance of this
        class or tuple of classes.
    '''

    # If one or more items of this iterable are *NOT* such instances...
    if not is_items_instance_of(iterable=iterable, cls=cls):
        # First such item.
        item_invalid = get_item_first_not_instance_of(
            iterable=iterable, cls=cls)

        # Raise an exception embedding this item.
        raise BetseIterableException(
            'Iterable item {} not instance of {}.'.format(item_invalid, cls))


@type_check
def die_unless_items_unique(iterable: IterableTypes) -> None:
    '''
    Raise an exception unless *all* items of the passed iterable are **unique**
    (i.e., no two distinct items are equal).

    Parameters
    ----------
    iterable: IterableTypes
        Iterable to be validated.

    Raises
    ----------
    BetseIterableException
        If at least one item of this iterable is a duplicate.
    '''

    # If one or more items of this iterable are duplicates...
    if not is_items_unique(iterable):
        # Set of all such duplicates.
        items_duplicate = get_items_duplicate(iterable)

        # Raise an exception embedding this set.
        raise BetseIterableException(
            'Iterable items {} duplicate.'.format(items_duplicate))

# ....................{ TESTERS                           }....................
@type_check
def is_reversible(iterable: IterableTypes) -> bool:
    '''
    `True` only if the passed iterable is **reversible** (i.e., successfully
    passable to the :func:`reversed` builtin).

    Specifically, this function returns `True` if this iterable either:

    * Defines the `__reversed__()` special method.
    * Defines the `__len__()` and `__getitem__()` special methods, satisfying
      the sequence protocol.

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be inspected.

    Returns
    ----------
    bool
        `True` only if this iterable is reversible.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj import objects

    # Return True only if this iterable either...
    return (
        # Defines the __reversed__() special method.
        objects.is_method(iterable, '__reversed__') or (
            # Defines the __len__() and __getitem__() special methods.
            objects.is_method(iterable, '__len__') and
            objects.is_method(iterable, '__getitem__')
        )
    )

# ....................{ TESTERS ~ items                   }....................
@type_check
def is_items_instance_of(iterable: IterableTypes, cls: TestableTypes) -> bool:
    '''
    ``True`` only if *all* items of the passed iterable are instances of the
    passed class or tuple of classes.

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be searched.
    cls : TestableTypes
        Class or tuple of classes to test that all items of this iterable be
        instances of.

    Returns
    ----------
    bool
        ``True`` only if *all* items of this iterable are instances of this
        class or tuple of classes.
    '''

    return all(isinstance(item, cls) for item in iterable)


@type_check
def is_items_unique(iterable: IterableTypes) -> bool:
    '''
    ``True`` only if *all* items of the passed iterable are **unique** (i.e.,
    no two distinct items are equal).

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be inspected.

    See Also
    ----------
    https://stackoverflow.com/a/5281641/2809027
        StackOverflow post strongly inspiring this implementation.

    Returns
    ----------
    bool
        ``True`` only if *all* items of this iterable are unique.
    '''

    # Set of all unique items of this iterable previously visited below.
    items_unique = set()

    # Return True only if no items in this iterable are duplicates (i.e., have
    # already been visited by a prior iteration of this test).
    return not any(
        # If this item is unique, add this item to this set as a side effect.
        item in items_unique or items_unique.add(item)
        for item in iterable)

# ....................{ TESTERS ~ item                    }....................
@type_check
def is_item_satisfying(
    iterable: IterableTypes, predicate: CallableTypes) -> bool:
    '''
    ``True`` only if some item of the passed iterable satisfies the passed
    **predicate** (i.e., callable accepting one parameter returning ``True``
    only if this parameter suffices).

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be searched.
    predicate : CallableTypes
        Callable accepting one parameter and returning ``True`` only if this
        parameter suffices.

    Returns
    ----------
    bool
        ``True`` only if some item of this iterable satisfies this predicate.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj.sentinels import SENTINEL

    # First item satisfying this predicate in this iterable if any *OR* the
    # sentinel placeholder otherwise.
    first_item = get_item_first_satisfying_or_sentinel(iterable, predicate)

    # Return True only if this item is *NOT* the sentinel, in which case some
    # item satisfies this predicate.
    return first_item is not SENTINEL


@type_check
def is_item_instance_of(
    iterable: IterableTypes, cls: TestableTypes) -> bool:
    '''
    ``True`` only if some item of the passed iterable is an instance of the
    passed class or tuple of classes.

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be searched.
    cls : TestableTypes
        Class or tuple of classes of the item to search this iterable for.

    Returns
    ----------
    bool
        ``True`` only if some item of this iterable is an instance of this
        class or tuple of classes.
    '''

    return is_item_satisfying(
        iterable=iterable, predicate=lambda item: isinstance(item, cls))

# ....................{ GETTERS                           }....................
@type_check
def get_items_duplicate(iterable: IterableTypes) -> set:
    '''
    Set of all duplicate items in the passed iterable.

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be inspected.

    Returns
    ----------
    set
        Set of all duplicate items in this iterable.

    See Also
    ----------
    https://stackoverflow.com/a/9835819/2809027
        Stackoverflow answer strongly inspiring this implementation.
    '''

    # Set of all unique items of this iterable previously visited below.
    items_unique = set()

    # Set of all duplicate items of this iterable to be returned.
    items_duplicate = set()

    # For each item of this iterable...
    for item in iterable:
        # If this item has been previously visited, this item is a duplicate.
        if item in items_unique:
            items_duplicate.add(item)
        # Else, this item is unique.
        else:
            items_unique.add(item)

    # Return this set of all duplicate items.
    return items_duplicate

# ....................{ GETTERS ~ first                   }....................
@type_check
def get_item_first(iterable: IterableTypes) -> object:
    '''
    First item non-destructively retrieved from the passed iterable if this
    iterable is non-empty *or* raise an exception otherwise (i.e., if this
    iterable is empty).

    If the passed iterable is a:

    * Sequence, this is guaranteed to be the first item of this sequence.
    * Non-sequence (e.g., :class:`set`, :class:`dict`), this should be assumed
      to be a random item. While most non-sequences guarantee predictable order
      of retrieval assuming no intervening changes, this is a fairly unreliable
      assumption.

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be inspected.

    Returns
    ----------
    object
        First element non-destructively retrieved from this iterable.

    Raises
    ----------
    BetseIterableException
        If this iterable is empty.

    See Also
    ----------
    https://stackoverflow.com/a/40054478/2809027
        Cecil's Stackoverflow answer strongly inspiring this implementation,
        complete with detailed timings of all alternative solutions.
    '''

    # If this iterable is empty, raise an exception. Since iteration of empty
    # iterables always succeeds, this condition must be manually tested
    # beforehand. Failing to do so would result Python raising the following
    # non-human-readable exception below:
    #
    #     NameError: name 'first_item' is not defined
    if not iterable:
        raise BetseIterableException('Iterable "{}" empty.'.format(iterable))

    # Yup! Shockingly, the most verbose and unwieldy solution is the fastest.
    # Break immediately after the first iteration of this iterable.
    first_item = None
    for first_item in iterable:
        break

    # Return the first element iterated above.
    return first_item

# ....................{ GETTERS ~ first : instance        }....................
@type_check
def get_item_first_instance_of(
    iterable: IterableTypes, cls: TestableTypes, **kwargs) -> object:
    '''
    First instance of the passed class or tuple of classes retrieved from the
    passed iterable if this iterable contains such an item *or* raise an
    exception otherwise (i.e., if this iterable contains no such item).

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be searched.
    cls : TestableTypes
        Class or tuple of classes of the item to search this iterable for.
    kwargs : dict
        Dictionary of all remaining keyword arguments to be passed as is to the
        :func:`get_item_first_satisfying` function.

    Returns
    ----------
    object
        First instance of this class in this iterable.

    Raises
    ----------
    BetseIterableException
        If this iterable contains no such item.

    See Also
    ----------
    :func:`get_item_first_satisfying_or_sentinel`
        Further details on ordering guarantees.
    '''

    return get_item_first_satisfying(
        iterable=iterable,
        predicate=lambda item: isinstance(item, cls),
        **kwargs
    )


@type_check
def get_item_first_not_instance_of(
    iterable: IterableTypes, cls: TestableTypes, **kwargs) -> object:
    '''
    First item of this iterable that is *not* an instance of the passed class
    or tuple of classes if this iterable contains such an item *or* raise an
    exception otherwise (i.e., if all items of this iterable are such
    instances).

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be searched.
    cls : TestableTypes
        Class or tuple of classes of the item to search this iterable for.
    kwargs : dict
        Dictionary of all remaining keyword arguments to be passed as is to the
        :func:`get_item_first_satisfying` function.

    Returns
    ----------
    object
        First non-instance of this class in this iterable.

    Raises
    ----------
    BetseIterableException
        If this iterable contains only such instances.

    See Also
    ----------
    :func:`get_item_first_satisfying_or_sentinel`
        Further details on ordering guarantees.
    '''

    return get_item_first_satisfying(
        iterable=iterable,
        predicate=lambda item: not isinstance(item, cls),
        **kwargs
    )

# ....................{ GETTERS ~ first : satisfying      }....................
@type_check
def get_item_first_satisfying_or_sentinel(
    iterable: IterableTypes, predicate: CallableTypes) -> object:
    '''
    First item of the passed iterable satisfying the passed **predicate**
    (i.e., callable accepting one parameter, returning ``True`` only if this
    parameter suffices) if this iterable contains such an item *or* the
    :attr:`betse.util.type.obj.sentinels.SENTINEL` placeholder constant
    otherwise.

    If the passed iterable is a:

    * Sequence, this is guaranteed to be the first such element.
    * Non-sequence (e.g., :class:`set`, :class:`dict`), this should be assumed
      to be a random such element. While most non-sequences guarantee
      predictable order of retrieval assuming no intervening changes, this is a
      fairly unreliable assumption.

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be inspected.
    predicate : CallableTypes
        Callable accepting one parameter and returning ``True`` only if this
        parameter suffices.

    Returns
    ----------
    object
        First element satisfying this predicate in this iterable if any *or*
        :attr:`betse.util.type.obj.sentinels.SENTINEL` otherwise.

    Raises
    ----------
    BetseIterableException
        If this iterable contains no such item.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj.sentinels import SENTINEL

    # Collective efficiency is our middle names.
    return next((item for item in iterable if predicate(item)), SENTINEL)


@type_check
def get_item_first_satisfying(
    iterable: IterableTypes,
    predicate: CallableTypes,
    exception_message: str = None,
) -> object:
    '''
    First item of the passed iterable satisfying the passed **predicate**
    (i.e., callable accepting one parameter, returning ``True`` only if this
    parameter suffices) if this iterable contains such an item *or* raise an
    exception otherwise (i.e., if this iterable contains no such item).

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be inspected.
    predicate : CallableTypes
        Callable accepting one parameter and returning ``True`` only if this
        parameter suffices.
    exception_message : optional[str]
        Exception message to be raised if no such item is found. Defaults to
        ``None``, in which case a general-purpose message is synthesized.

    Returns
    ----------
    object
        First item satisfying this predicate in this iterable.

    Raises
    ----------
    BetseIterableException
        If this iterable contains no such item.

    See Also
    ----------
    :func:`get_item_first_satisfying_or_sentinel`
        Further details on ordering guarantees.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj.sentinels import SENTINEL

    # First item satisfying this predicate in this iterable if any *OR* the
    # sentinel placeholder otherwise.
    first_item = get_item_first_satisfying_or_sentinel(iterable, predicate)

    # If no item satifies this predicate, raise an exception.
    if first_item is SENTINEL:
        # If no exception message is passed, synthesize a default message.
        if exception_message is None:
            exception_message = (
                'Iterable "{}" item satisfying predicate {} not found.'.format(
                    iterable, predicate))

        # Raise this exception.
        raise BetseIterableException(exception_message)

    # Else, return this element.
    return first_item

# ....................{ GETTERS ~ last : instance         }....................
@type_check
def get_item_last_instance_of(
    iterable: IterableTypes, cls: TestableTypes, **kwargs) -> object:
    '''
    Last instance of the passed class or tuple of classes retrieved from the
    passed iterable if this iterable contains such an item *or* raise an
    exception otherwise (i.e., if this iterable contains no such item).

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be searched.
    cls : TestableTypes
        Class or tuple of classes of the item to search this iterable for.
    kwargs : dict
        Dictionary of all remaining keyword arguments to be passed as is to the
        :func:`get_item_last_satisfying` function.

    Returns
    ----------
    object
        Last instance of this class in this iterable.

    Raises
    ----------
    BetseIterableException
        If this iterable contains no such item.

    See Also
    ----------
    :func:`get_item_last_satisfying`
        Further details on ordering guarantees.
    '''

    return get_item_last_satisfying(
        iterable=iterable,
        predicate=lambda item: isinstance(item, cls),
        **kwargs
    )


@type_check
def get_item_last_instance_of_or_none(
    iterable: IterableTypes, cls: TestableTypes, **kwargs) -> object:
    '''
    Last instance of the passed class or tuple of classes retrieved from the
    passed iterable if this iterable contains such an item *or* ``None``
    otherwise (i.e., if this iterable contains no such element).

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be searched.
    cls : TestableTypes
        Class or tuple of classes of the item to search this iterable for.
    kwargs : dict
        Dictionary of all remaining keyword arguments to be passed as is to the
        :func:`get_item_last_satisfying` function.

    Returns
    ----------
    object
        Last instance of this class in this iterable if any *or*
        :attr:`betse.util.type.obj.sentinels.SENTINEL` otherwise.

    Raises
    ----------
    BetseIterableException
        If this iterable contains no such item.

    See Also
    ----------
    :func:`get_item_last_satisfying`
        Further details on ordering guarantees.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj.sentinels import SENTINEL

    # Last instance of this class in this iterable if any or the sentinel
    # placeholder otherwise.
    item_found = get_item_last_satisfying_or_sentinel(
        iterable=iterable,
        predicate=lambda item: isinstance(item, cls),
        **kwargs
    )

    # Return this instance if *NOT* the sentinal placeholder or None otherwise.
    return item_found if item_found is not SENTINEL else None

# ....................{ GETTERS ~ last : satisfying      }....................
@type_check
def get_item_last_satisfying_or_sentinel(
    iterable: IterableTypes, predicate: CallableTypes) -> object:
    '''
    Last element of the passed iterable satisfying the passed **predicate**
    (i.e., callable accepting one parameter, returning ``True`` only if this
    parameter suffices) if this iterable contains such an element *or* the
    :attr:`betse.util.type.obj.sentinels.SENTINEL` placeholder constant
    otherwise.

    If the passed iterable is a:

    * Sequence, this is guaranteed to be the last such element.
    * Non-sequence (e.g., :class:`set`, :class:`dict`), this should be assumed
      to be a random such element. While most non-sequences guarantee
      predictable order of retrieval assuming no intervening changes, this is a
      fairly unreliable assumption.

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be inspected.
    predicate : CallableTypes
        Callable accepting one parameter and returning ``True`` only if this
        parameter suffices.

    Returns
    ----------
    object
        Last element satisfying this predicate in this iterable if any *or*
        :attr:`betse.util.type.obj.sentinels.SENTINEL` otherwise.

    Raises
    ----------
    BetseIterableException
        If this iterable contains no such element.
    '''

    # For simplicity, the existing get_item_first_satisfying_or_sentinel()
    # function is deferred to by returning the first element in the reverse of
    # this iterable satisfying this predicate.
    return get_item_first_satisfying_or_sentinel(
        # For safety, this iterable is reversed via the high-level reverse()
        # function rather than the low-level reversed() builtin; the latter
        # fails to generically support all possible iterable types.
        iterable=reverse(iterable),
        predicate=predicate,
    )



@type_check
def get_item_last_satisfying(
    iterable: IterableTypes,
    predicate: CallableTypes,
    exception_message: str = None,
) -> object:
    '''
    Last element of the passed iterable satisfying the passed **predicate**
    (i.e., callable accepting one parameter, returning ``True`` only if this
    parameter suffices) if this iterable contains such an element *or* raise an
    exception otherwise (i.e., if this iterable contains no such element).

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be inspected.
    predicate : CallableTypes
        Callable accepting one parameter and returning ``True`` only if this
        parameter suffices.
    exception_message : optional[str]
        Exception message to be raised if no such element is found. Defaults to
        ``None``, in which case a suitably general-purpose message is
        synthesized.

    Returns
    ----------
    object
        Last element satisfying this predicate in this iterable.

    Raises
    ----------
    BetseIterableException
        If this iterable contains no such element.

    See Also
    ----------
    :func:`get_item_first_satisfying_or_sentinel`
        Further details on ordering guarantees.
    '''

    # For simplicity, the existing get_item_first_satisfying() function is
    # deferred to by returning the first element in the reverse of this
    # iterable satisfying this predicate.
    return get_item_first_satisfying(
        # For safety, this iterable is reversed via the high-level reverse()
        # function rather than the low-level reversed() builtin; the latter
        # fails to generically support all possible iterable types.
        iterable=reverse(iterable),
        predicate=predicate,
        exception_message=exception_message,
    )

# ....................{ CONVERTERS                        }....................
@type_check
def to_iterable(iterable: IterableTypes, cls: ClassType) -> IterableTypes:
    '''
    Convert the passed iterable into an iterable of the passed type.

    If this iterable is:

    * Of the same type as the passed type, this iterable is returned as is.
    * A non-Numpy iterable (e.g., :class:`list`) and the passed type is that
      of:
      * Another non-Numpy iterable (e.g., :class:`tuple`), this iterable is
        converted into an instance of this type. To do so, this type's
        ``__init__`` method is expected to accept this :class:`list` as a
        single positional argument.
      * A Numpy array, this iterable is converted to a Numpy array via the
        :func:`betse.lib.numpy.nparray.from_iterable` function.
    * A Numpy array, this array is converted to the passed type via the
      :func:`betse.lib.numpy.nparray.to_iterable` function.

    Parameters
    ----------
    iterable: IterableTypes
        Source iterable to be converted.
    cls : ClassType
        Type of the target iterable to convert this source iterable into.

    Returns
    ----------
    IterableTypes
        Target iterable converted from this source iterable.
    '''

    # Avoid importing third-party packages at the top level, for safety.
    from betse.lib.numpy import nparray
    from numpy import ndarray

    # Type of the source iterable.
    iterable_src_type = type(iterable)

    # If the source and target iterables are of the same type, return this
    # source iterable as is.
    if iterable_src_type is cls:
        return iterable

    # Else if the source iterable is a Numpy array, defer to logic elsewhere.
    if iterable_src_type is ndarray:
        return nparray.to_iterable(array=iterable, cls=cls)

    # Else if the target iterable is a Numpy array, defer to logic elsewhere.
    if cls is ndarray:
        return nparray.from_iterable(iterable)

    # Else, the source iterable is a non-Numpy iterable. Defer to the
    # constructor of the target iterable for conversion.
    return cls(iterable)

# ....................{ CONSUMERS                         }....................
@type_check
def consume(iterable: IterableTypes, iterations: int) -> object:
    '''
    Consume the passed number of iterations from the passed iterable by
    advancing this iterable forward by this number of iterations.

    For efficiency, this iterable is consumed at C speeds through standard
    objects implemented in low-level C rather than high-level Python.

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be consumed.
    iterations : int
        Number of iterations to advance this iterable. This number should be
        strictly positive (i.e., ``iterations >= 1``).

    Returns
    ----------
    object
        Object yielded by the last iterable iteration (i.e., the last
        ``next()`` method called on this iterable) if any *or* ``None`` if this
        iterable was already exhausted (i.e., empty) when passed.

    See Also
    ----------
    http://docs.python.org/3/library/itertools.html
        Official documentation strongly inspiring this function.
    '''
    assert types.is_int_positive(iterations), (
        types.assert_not_int_positive(iterations))

    # Advance to the empty slice starting (and hence immediately ending) at the
    # passed iterable position and return the value of this iterable at this
    # position if any.
    return next(slice(iterable, iterations, iterations), None)


@type_check
def exhaust(iterable: IterableTypes) -> object:
    '''
    Exhaust the passed iterable by advancing this iterable directly past its
    last iteration.

    For efficiency, this iterable is consumed at C speeds through standard
    objects implemented in low-level C rather than high-level Python.

    Caveats
    ----------
    **This function should only be called for finite iterables.** If the passed
    iterable:

    * Explicitly halts with a :class:`StopIteration` exception and hence is
      finite, this function exhausts this iterable as expected.
    * Does *not* explicitly halt with a :class:`StopIteration` exception and
      hence is infinite, this function reduces to an **infinite loop.**

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be exhausted.

    Returns
    ----------
    object
        Object yielded by the last iterable iteration (i.e., the last
        ``next()`` method called on this iterable) if any *or* ``None`` if this
        iterable was already exhausted (i.e., empty) when passed.

    See Also
    ----------
    http://docs.python.org/3/library/itertools.html
        Official documentation strongly inspiring this function.
    '''

    # For efficiency, feed this iterable into a zero-length deque retaining
    # only the last iterated value if any.
    iterable_deque = deque(iterable, maxlen=1)

    # If this iterable was *NOT* already exhausted, return its last value.
    if iterable_deque:
        return iterable_deque[0]
    # Else, this iterable was already exhausted. Return nothing.
    else:
        return None

# ....................{ INVERTERS                         }....................
@type_check
def invert_iterable_unique(iterable: IterableTypes) -> MappingType:
    '''
    Dictionary inverted from the passed iterable if **internally unique**
    (i.e., containing no duplicate items) *or* raise an exception otherwise.

    Specifically:

    * If this iterable is a dictionary, the
      :func:`betse.util.type.mapping.mappings.invert_dict_unique` function is
      silently deferred to. The type of the returned dictionary is guaranteed to
      be the same as the type of the passed dictionary.
    * Else, the returned dictionary maps from each item of this iterable to the
      0-based index of that item in this iterable. The type of this dictionary
      is guaranteed to *not* be the same as the type of this iterable.

    Parameters
    ----------
    iterable : IterableTypes
        Internally unique iterable to be inverted.

    Returns
    ----------
    MappingType
        Dictionary inverted from this iterable as detailed above.

    Raises
    ----------
    BetseIterableException
        If at least one item of this iterable is a duplicate.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.mapping import mappings

    # If this iterable is a mapping...
    if mappings.is_mapping(iterable):
        # Invert this iterable specifically as a mapping.
        #
        # While mappings are technically iterables and hence invertable via the
        # generic approach applied below, doing so incorrectly returns a
        # dictionary mapping from the keys of this dictionary to arbitrary
        # 0-based integers. As the adjective "arbitrary" implies, this renders
        # the resulting dictionary effectively useless for most purposes.
        return mappings.invert_map_unique(iterable)
    # Else, this iterable is a non-mapping. In this case, a generic approach
    # suffices... usually.
    else:
        # If one or more items of this iterable are duplicates, raise an
        # exception.
        die_unless_items_unique(iterable)

        # One-liners for Great Glory.
        return {item: item_index for item_index, item in enumerate(iterable)}

# ....................{ ITERATORS                         }....................
@type_check
def iter_items(*iterables: IterableTypes) -> GeneratorType:
    '''
    Generator yielding each item of each of the passed iterables (in both the
    internal order of each iterable and the passed order of iterables),
    effectively "chaining" these iterables together.

    This function is preferable for pure iteration over multiple iterables, in
    which case a composite iterable of the same type is *not* required.

    Parameters
    ----------
    iterables : tuple[IterableTypes]
        Tuple of all iterables whose items are to be iterated over.

    Yields
    ----------
    object
        Current item of the current iterable.

    See Also
    ----------
    :func:`join`
        Less efficient alternative producing a composite iterable.
    https://stackoverflow.com/a/14342360/2809027
        Stackoverflow answer strongly inspiring this implementation.

    Examples
    ----------
        >>> from betse.util.type import iterables
        >>> in_xanadu = ('did', 'Kubla', 'Khan')
        >>> a_stately = ('pleasure-dome', 'decree:')
        >>> for word in iterables.iter_items(in_xanadu, a_stately):
        ...     print(word, end=' ')
        did Kubla Khan pleasure-dome decree:
    '''

    # Pure Python 3.x. Join the party.
    #
    # Note that the itertools.chain.from_iterable() could also be called here,
    # but that doing so is less Pythonic than the current approach.
    for iterable in iterables:
        yield from iterable

# ....................{ JOINERS                           }....................
@type_check
def join(*iterables: IterableTypes) -> IterableTypes:
    '''
    Join each item of each of the passed iterables (in both the internal order
    of each iterable and the passed order of iterables) into a new iterable of
    the same type as the first such iterable.

    Parameters
    ----------
    iterables : tuple[IterableTypes]
        Tuple of all iterables whose items are to be joined together.

    Returns
    ----------
    IterableTypes
        Iterable of the same type as the first passed iterable containing each
        item of each of the passed iterables.

    Raises
    ----------
    BetseIterableException
        If the type of the first passed iterable is :class:`str`, in which case
        the builtin :func:`str.join` function should be called for both
        efficiency and sanity instead.

    See Also
    ----------
    :func:`iter_items`
        More efficient alternative intended for pure iteration, in which case a
        composite iterable of the same type is *not* required.
    https://stackoverflow.com/a/14342360/2809027
        Stackoverflow answer strongly inspiring this implementation.

    Examples
    ----------
        >>> from betse.util.type import iterables
        >>> where_alph = ('the', 'sacred', 'river,', 'ran')
        >>> through_caverns = ('measureless', 'to', 'man')
        >>> for word in iterables.iter_items(where_alph, through_caverns):
        ...     print(word, end=' ')
        the sacred river, ran measureless to man
    '''

    # Iterator over each of the passed iterables.
    iterator = iter(iterables)

    # First passed iterable.
    iterable_first = next(iterator)

    # Type of this iterable.
    iterable_first_type = type(iterable_first)

    # If this iterable is a string, raise an exception.
    if iterable_first_type is str:
        raise BetseIterableException(
            'String "{}" not joinable by iterables.join(). '
            'Consider calling str.join() instead.'.format(iterable_first))

    # Return a new iterable of this type over each item of each passed iterable.
    return iterable_first_type(iter_items(*iterables))

# ....................{ REVERSERS                         }....................
@type_check
def reverse(iterable: IterableTypes) -> IterableTypes:
    '''
    Reverse the passed iterable into a new iterable of differing type
    containing all elements of the passed iterable in reverse order.

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be returned reversed. For generality, this iterable is
        *not* modified by this function.

    Returns
    ----------
    IterableTypes
        Iterable reversed from the passed iterable. For efficiency, this
        iterable is only a shallow rather than deep copy of the passed
        iterable.
    '''

    # If this iterable is *NOT* reversible as is, convert this iterable into
    # the most space- and time-efficient iterable containing the same elements
    # that *IS* reversible -- in this case, a tuple.
    if not is_reversible(iterable):
        iterable = tuple(iterable)

    # Return the result of the efficient reversed() builtin on this iterable,
    # now guaranteed to be reversible as is.
    return reversed(iterable)

# ....................{ SORTERS ~ ascending               }....................
@type_check
def sort_ascending(iterable: IterableTypes) -> IterableTypes:
    '''
    Iterable sorted from the passed iterable in ascending order.

    Each element of this iterable is compared to each other element of this
    iterable via the ``<`` operator, implicitly calling the ``__le__()``
    special method of these elements. Each element is ideally but *not*
    necessarily of the same type. If each element is:

    * A string, these strings are sorted in **ascending lexicographic order**
      (i.e., traditional order of dead-tree dictionaries and encyclopedias).
    * A number (i.e., integer or a float), these numbers are sorted in
      **ascending numeric order.**

    Parameters
    ----------
    iterable : IterableTypes
        Unsorted iterable to be returned sorted. For generality, this iterable
        is *not* modified by this function.

    Returns
    ----------
    IterableTypes
        Iterable sorted from and of the same type as the passed iterable. For
        efficiency, this iterable is only a shallow rather than deep copy of
        the passed iterable. Note lastly that the class of the passed iterable
        *must* define an ``__init__()`` method accepting a list.
    '''

    # Type of the passed iterable.
    iterable_type = type(iterable)

    # Return an iterable of the same type, converted from the sorted list
    # returned by the sorted() builtin.
    return iterable_type(sorted(iterable))


@type_check
def sort_by_index_ascending(
    iterable: IterableTypes, subiterable_index: object) -> IterableTypes:
    '''
    Iterable of subiterables sorted from the passed iterable of subiterables in
    ascending order of the value of each element at the passed key or index of
    each subiterable of this iterable.

    Each element at the passed key or index of each subiterable of this
    iterable is compared to each other element at each other key or index of
    each other subiterable of this iterable via the ``<`` operator, implicitly
    calling the ``__le__()`` special method of these elements. Each element is
    ideally but *not* necessarily of the same type. If each element is:

    * A string, these strings are sorted in **ascending lexicographic order**
      (i.e., traditional order of dead-tree dictionaries and encyclopedias).
    * A number (i.e., integer or a float), these numbers are sorted in
      **ascending numeric order.**

    Parameters
    ----------
    iterable : IterableTypes
        Unsorted iterable of subiterables to be returned sorted. For
        generality, neither this iterable nor these subiterables are modified
        by this function.
    subiterable_index : object
        Object with which to index each subiterable of this iterable. The type
        of this object _must_ be a type accepted by the ``__getitem__()``
        special method of each subiterable. Specifically, if each subiterable
        is a:

        * **Mapping** (e.g., :class:`dict`), this object *must* be hashable.
        * **Sequence** (e.g., :class:`list`, :class:`tuple`), this object
          *must* be either:

          * An integer.
          * A :func:`slice` object.

    Returns
    ----------
    IterableTypes
        Iterable of subiterables sorted from and of the same type as the passed
        iterable of subiterables. For efficiency, this iterable is only a
        shallow rather than deep copy of the passed iterable. Note lastly that
        the class of the passed iterable *must* define an ``__init__()`` method
        accepting a list.
    '''

    # Type of the passed iterable.
    iterable_type = type(iterable)

    # Return an iterable of the same type, converted from the sorted list
    # returned by the sorted() builtin.
    #
    # For efficiency, elements of each subiterable of this iterable are
    # retrieved via the function internally created and called by calling an
    # instance of the standard "itemgetter" class. While technically an
    # instance of this class, this class is intended to be treated as a simple
    # function passed the desired index returning a function passed a
    # subiterable returning the element at that index: e.g.,
    #
    #     >>> penguins = [('adelie', 0xFEEDFACE), ('gentoo', 0xDEADBEEF)]
    #     >>> penguin_id = itemgetter(2)
    #     >>> penguin_id(penguins[0]) == 0xFEEDFACE
    #     True
    #
    # Despite the internal complexity and object overhead imposed by the
    # "itemgetter" class, this approach has been definitively profiled to be
    # 126% faster on average than the traditional
    # "lambda subiterable: subiterable[subiterable_index]" approach --
    # presumably due to hidden overhead imposed by capturing all local
    # variables into the closure context encapsulated by the lambda. See also
    # the following Stackoverflow answer exhibiting this profiling:
    #
    #     https://stackoverflow.com/a/17243726/2809027
    return iterable_type(sorted(iterable, key=itemgetter(subiterable_index)))

# ....................{ SORTERS ~ descending              }....................
@type_check
def sort_descending(iterable: IterableTypes) -> IterableTypes:
    '''
    Iterable sorted from the passed iterable in descending order.

    Each element of this iterable is compared to each other element of this
    iterable via the ``>`` operator, implicitly calling the ``__ge__()``
    special method of these elements. Each element is ideally but *not*
    necessarily of the same type. If each element is:

    * A string, these strings are sorted in **descending lexicographic order**
      (i.e., reverse order of dead-tree dictionaries and encyclopedias).
    * A number (i.e., integer or a float), these numbers are sorted in
      **descending numeric order.**

    See Also
    ----------
    :func:`sort_ascending`
        Further details.
    '''

    # Type of the passed iterable.
    iterable_type = type(iterable)

    # Return an iterable of the same type, converted from the sorted list
    # returned by the sorted() builtin.
    return iterable_type(sorted(iterable, reverse=True))


@type_check
def sort_by_index_descending(
    iterable: IterableTypes, subiterable_index: object) -> IterableTypes:
    '''
    Iterable of subiterables sorted from the passed iterable of subiterables in
    descending order of the value of each element at the passed key or index of
    each subiterable of this iterable.

    Each element at the passed key or index of each subiterable of this
    iterable is compared to each other element at each other key or index of
    each other subiterable of this iterable via the ``>`` operator, implicitly
    calling the ``__ge__()`` special method of these elements. Each element is
    ideally but *not* necessarily of the same type. If each element is:

    * A string, these strings are sorted in **descending lexicographic order**
      (i.e., reverse order of dead-tree dictionaries and encyclopedias).
    * A number (i.e., integer or a float), these numbers are sorted in
      **descending numeric order.**

    See Also
    ----------
    :func:`sort_by_index_ascending`
        Further details.
    '''

    # Type of the passed iterable.
    iterable_type = type(iterable)

    # Return an iterable of the same type, converted from the sorted list
    # returned by the sorted() builtin. See sort_by_index_descending() for
    # commentary, particularly on the "key" parameter.
    return iterable_type(sorted(
        iterable, key=itemgetter(subiterable_index), reverse=True))

# ....................{ ZIPPERS                           }....................
#FIXME: Unit test us up.
@type_check
def zip_isometric(*iterables: IterableTypes) -> GeneratorType:
    '''
    Generator zipping all passed iterables required to be of the same length.

    This generator iteratively yields an n-tuple, where:

    * n is the length of each passed iterable.
    * The i-th element of this tuple is in the i-th passed iterable.

    Parameters
    ----------
    iterables : IterableTypes
        Tuple of iterables of the same length to be zipped.

    Returns
    ----------
    GeneratorType
        Generator zipping these iterables.

    Raises
    ----------
    BetseIterableException
        If any passed iterable differs in length from any other passed
        iterable.

    See Also
    ----------
    https://stackoverflow.com/a/32954700/2809027
        Stackoverflow answer strongly inspiring this implementation.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj.sentinels import SENTINEL

    # Iteratively zip and yield each n-tuple from the passed n iterables. To
    # efficiently detect iterables of insufficient length, the C-based
    # zip_longest() function is called to fill all iterables of insufficient
    # length with sentinel objects to the expected length.
    #
    # After zipping but before yielding each n-tuple, this n-tuple is then
    # manually searched for sentinel objects. Since these objects may reside at
    # any index of this n-tuple, the entire n-tuple *MUST* be searched in an
    # O(n) manner. While unfortunate, this approach remains substantially more
    # efficient than all alternatives -- largely due to the efficacy of the
    # zip_longest() function. Surprisingly, this "for" loop-based approach has
    # been timed to be faster than the following generator expression:
    #
    #     return (
    #         ntuple if SENTINEL not in ntuple else (
    #             _zip_isometric_error(iterables, ntuple))
    #         for ntuple in zip_longest(
    #             *iterables, fillvalue=SENTINEL)
    #     )
    for ntuple in itertools.zip_longest(
        *iterables, fillvalue=SENTINEL):
        # If this n-tuple contains a sentinel, at least one passed iterable is
        # of insufficient length. Raise a human-readable exception indicating
        # the index and contents of this iterable.
        if SENTINEL in ntuple:
            raise _zip_isometric_error(iterables, ntuple)

        # Else, this n-tuple is valid. Yield it up!
        yield ntuple


def _zip_isometric_error(iterables: tuple, ntuple: tuple) -> None:
    '''
    Raise an exception indicating that the iterable of the passed tuple of
    iterables identified by the passed `n`-tuple is of smaller length than
    other iterables in this tuple of iterables.

    This private function is *only* intended to be called by the
    :func:`zip_isometric` function.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj.sentinels import SENTINEL

    # Index of the erroneously short iterable in this tuple of iterables,
    # identical to the index of the first sentinel in the passed zipped tuple.
    iterable_short_index = None
    for iterable_short_index, item in enumerate(ntuple):
        if item is SENTINEL:
            break

    # This erroneously short iterable.
    iterable_short = iterables[iterable_short_index]

    # If the first iterable is of predefined length (e.g., is *NOT* a generator
    # of dynamic length), end this exception message with this length.
    if isinstance(iterables[0], SizedType):
        exception_suffix = 'length {} of prior iterables'.format(
            len(iterables[0]))
    # Else, end this exception message as is.
    else:
        exception_suffix = 'length of prior iterables'

    # If this erroneously short iterable is of predefined length (e.g., is
    # *NOT* a generator of dynamic length), begin this exception message with
    # this length and end this message with this iterable's contents.
    if isinstance(iterable_short, SizedType):
        exception_prefix = "Length {} of iterable {}".format(
            len(iterable_short), iterable_short_index)
        exception_suffix += ': {!r}'.format(iterable_short)
    # Else, begin this exception message as is.
    else:
        exception_prefix = "Length of iterable {}".format(iterable_short_index)

    # Raise this exception.
    raise BetseIterableException('{} differs from {}'.format(
        exception_prefix, exception_suffix))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Export-specific enumeration types** (i.e., :class:`enum.Enum` subclasses
expressing mutually exclusive choices specific to post-simulation exports).
'''

# ....................{ IMPORTS                            }....................
from betse.util.type.enums import make_enum

# ....................{ ENUMS                              }....................
SimExportType = make_enum(
    class_name='SimExportType', member_names=('ANIM', 'CSV', 'PLOT',))
'''
Enumeration of all supported types of **post-simulation exports** (e.g., plots,
animations, comma-separated value (CSV) files).

Attributes
----------
ANIM : enum
    Post-simulation animations, visualizing the cell cluster across one or more
    sampled time steps of an initialization or simulation phase as the same
    number of frames of an animated video.
CSV : enum
    Post-simulation comma-separated value (CSV) files, aggregating raw data for
    the cell cluster across one or more sampled time steps of an initialization
    or simulation phase.
PLOT : enum
    Post-simulation plots, visualizing the cell cluster across one or more
    sampled time steps of an initialization or simulation phase as a single
    non-animated image.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests for the :func:`betse.util.py.modules.is_c_extension` function.
'''

# ....................{ IMPORTS                            }....................
# from pytest import fixture
# import pytest, sys

# ....................{ TESTS                              }....................
def test_is_c_extension_unmonkeypatched() -> None:
    '''
    Unit test the :func:`betse.util.py.modules.is_c_extension` function
    _without_ monkeypatching the PEP 302-specific `__loader__` attribute of
    tested modules.
    '''

    # Imports deferred for safety.
    from betse.util.type import modules
    from numpy.core import multiarray

    # Ensure that a C extension is correctly detected as such. Testing this
    # particular C extension is critical, as the core codebase explicitly
    # performs the same call at application startup.
    assert modules.is_c_extension(multiarray) is True

    # Ensure that a pure-Python submodule is *NOT* detected to be a C extension.
    assert modules.is_c_extension(modules) is False


def test_is_c_extension_monkeypatched(monkeypatch) -> None:
    '''
    Test all calls of the :func:`betse.util.py.modules.is_c_extension` function
    by monkeypatching the PEP 302-specific `__loader__` attribute of tested
    modules.

    This monkeypatch improves code coverage by exercising an infrequently used
    fallback codepath in this function.

    Parameters
    ----------
    monkeypatch : MonkeyPatch
        Builtin fixture object permitting object attributes to be safely
        modified for the duration of this unit test.
    '''

    # Imports deferred for safety.
    from betse.util.type import modules
    from numpy.core import multiarray

    # Remove the the PEP 302-specific "__loader__" attribute of each such
    # module *BEFORE* performing testing. Since this attribute need *NOT* exist,
    # prevent exceptions from being raised when this attribute does *NOT* exist.
    for module_object in (modules, multiarray):
        monkeypatch.delattr(module_object, name='__loader__', raising=False)

    # Defer to the existing base unit test.
    test_is_c_extension_unmonkeypatched()



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **application dependency** (i.e., both mandatory and optional
third-party Python packages required by this application) facilities.

This low-level submodule defines functions intended to be called by high-level
submodules (e.g., :mod:`betse.util.cli.cliabc`) *before* attempting to import
any such dependencies.
'''

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable exceptions on missing mandatory dependencies,
# the top-level of this module may import *ONLY* from packages guaranteed to
# exist at initial runtime (i.e., standard Python and application packages).
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

from betse import metadata, metadeps
from betse.exceptions import BetseLibException
from betse.util.io.log import logs
from betse.util.type.types import type_check, MappingType, StrOrNoneTypes

# ....................{ GLOBALS                           }....................
_IS_INITTED = False
'''
``True`` only if the :func:`init` function has already been called.

That function uses this private boolean to guard against repeated invocations
of the :func:`init` function from multiple modules in the same Python process
(e.g., :mod:`betse.science.__init__`, :mod:`betse.util.cli.cliabc`). While that
function does technically support repeated calls, each additional call after
the first inefficiently performs no meaningful work and is thus safely
ignorable.
'''

# ....................{ EXCEPTIONS                        }....................
def die_unless_runtime_mandatory_all() -> None:
    '''
    Raise an exception unless all mandatory runtime dependencies of this
    application are **satisfiable** (i.e., both importable and of a
    satisfactory version) *and* all external commands required by these
    dependencies (e.g., GraphViz's ``dot`` command) reside in the current
    ``${PATH}``.

    Equivalently, this function raises an exception if at least one such
    dependency is unsatisfied. For importable unsatisfied dependencies with
    :mod:`setuptools`-specific metadata (e.g., ``.egg-info/``-suffixed
    subdirectories of the ``site-packages/`` directory for the active Python 3
    interpreter, typically created by :mod:`setuptools` at install time), this
    function additionally validates the versions of these dependencies to
    satisfy all application requirements.

    Raises
    ----------
    BetseLibException
        If at least one mandatory runtime dependency is unsatisfiable.
    '''

    die_unless_requirements_dict(metadeps.RUNTIME_MANDATORY)


@type_check
def die_unless_runtime_optional(*requirement_names: str) -> None:
    '''
    Raise an exception unless all optional runtime dependencies of this
    application with the passed :mod:`setuptools`-specific project names are
    **satisfiable** (i.e., both importable and of a satisfactory version)
    *and* all external commands required by these dependencies (e.g.,
    GraphViz's ``dot`` command) reside in the current ``${PATH}``.

    Parameters
    ----------
    requirement_names : Tuple[str]
        Tuple of the names of all :mod:`setuptools`-specific projects
        implementing these dependencies (e.g., ``NetworkX``). If any such name
        is unrecognized (i.e., is *not* a key of the
        :data:`metadeps.RUNTIME_OPTIONAL` dictionary), an exception is raised.

    Raises
    ----------
    BetseLibException
        If at least one such dependency is unsatisfiable.

    See Also
    ----------
    :func:`die_unless_runtime_mandatory_all`
        Further details.
    '''

    die_unless_requirements_dict_keys(
        metadeps.RUNTIME_OPTIONAL, *requirement_names)

# ....................{ EXCEPTIONS ~ dict                 }....................
@type_check
def die_unless_requirements_dict(requirements_dict: MappingType) -> None:
    '''
    Raise an exception unless all dependencies described by the passed
    dictionary are **satisfiable** (i.e., both importable and of a satisfactory
    version) *and* all external commands required by these dependencies (e.g.,
    GraphViz's ``dot`` command) reside in the current ``${PATH}``.

    Raises
    ----------
    BetseLibException
        If at least passed dependency is unsatisfiable.
    '''

    # Avoid circular import dependencies.
    from betse.util.type import modules

    # If the "pkg_resources" setuptools dependency is missing, raise an
    # exception *BEFORE* importing this dependency below.
    modules.die_unless_module(
        module_name='pkg_resources',
        exception_message='Mandatory dependency "pkg_resources" not found.',
    )

    # Import this submodule *AFTER* validating "pkg_resources" to exist above.
    from betse.lib.setuptools import setuptool

    # Validate these dependencies.
    setuptool.die_unless_requirements_dict(requirements_dict)

    # Validate all external commands required by these dependencies.
    die_unless_commands(*requirements_dict.keys())


@type_check
def die_unless_requirements_dict_keys(
    requirements_dict: MappingType, *requirement_names: str) -> None:
    '''
    Raise an exception unless all dependencies with the passed
    :mod:`setuptools`-specific project names described by the passed dictionary
    are **satisfiable** (i.e., both importable and of a satisfactory version)
    *and* all external commands required by these dependencies (e.g.,
    GraphViz's ``dot`` command) reside in the current ``${PATH}``.

    Parameters
    ----------
    requirements_dict : MappingType
        Dictionary mapping from the names of all :mod:`setuptools`-specific
        projects implementing these dependencies to the requirements strings
        constraining these dependencies.
    requirement_names : tuple[str]
        Tuple of the names of all :mod:`setuptools`-specific projects
        implementing these dependencies (e.g., ``NetworkX``). If any such
        name is *not* a key of this dictionary, an exception is raised.

    Raises
    ----------
    BetseLibException
        If at least passed dependency is unsatisfiable.
    '''

    # Avoid circular import dependencies.
    from betse.lib.setuptools import setuptool

    # Validate these dependencies.
    setuptool.die_unless_requirements_dict_keys(
        requirements_dict, *requirement_names)

    # Validate all external commands required by these dependencies.
    die_unless_commands(*requirement_names)

# ....................{ EXCEPTIONS ~ commands             }....................
@type_check
def die_unless_commands(*requirement_names: str) -> None:
    '''
    Raise an exception unless all external commands required by all application
    dependencies (of any type, including optional, mandatory, runtime, testing,
    or otherwise) with the passed :mod:`setuptools`-specific project names are
    **installed** (i.e., are executable files in the current ``${PATH}``).

    Parameters
    ----------
    requirement_names : Tuple[str]
        Tuple of the names of all :mod:`setuptools`-specific projects
        corresponding to these dependencies (e.g., ``NetworkX``).

    Raises
    ----------
    BetseLibException
        If any external command required by any such dependency is *not* found.
    '''

    # Avoid circular import dependencies.
    from betse.util.path.command import cmdpath

    # If any external command required by any such dependency is *NOT* found,
    # iteratively search for the first such missing command and raise a
    # human-readable exception synopsizing this command. For efficiency, this
    # inefficient iteration is performed *ONLY* as required.
    if not is_commands(*requirement_names):
        # For the name of each such dependency...
        for requirement_name in requirement_names:
            # For each "betse.metadata.DependencyCommand" instance
            # describing each external command required by this dependency if
            # any *OR* the empty tuple otherwise...
            for dependency_command in metadeps.EXTERNAL_COMMANDS.get(
                requirement_name, ()):
                # If this command is *NOT* in the ${PATH}, raise an exception.
                if not cmdpath.is_pathable(dependency_command.basename):
                    raise BetseLibException(
                        'Dependency "{}" unsatisfied, as '
                        '{} not installed '
                        '(i.e., command "{}" not found in ${{PATH}}).'.format(
                            requirement_name,
                            dependency_command.name,
                            dependency_command.basename,
                        ))

# ....................{ TESTERS                           }....................
@type_check
def is_commands(*requirement_names: str) -> bool:
    '''
    ``True`` only if all external commands required by all application
    dependencies (of any type, including optional, mandatory, runtime, testing,
    or otherwise) with the passed :mod:`setuptools`-specific project names are
    **installed** (i.e., are executable files in the current ``${PATH}``).

    Parameters
    ----------
    requirement_names : Tuple[str]
        Tuple of the names of all :mod:`setuptools`-specific projects
        corresponding to these dependencies (e.g., ``NetworkX``).
    '''

    # Avoid circular import dependencies.
    from betse.util.path.command import cmdpath

    # Return True only if...
    return all(
        # Each external command required by each dependency is in the ${PATH}.
        cmdpath.is_pathable(dependency_command.basename)
        # For the name of each passed dependency...
        for requirement_name in requirement_names
        # For the tuple of all "betse.metadata.DependencyCommand" instances
        # describing all external commands required by this dependency if any
        # *OR* the empty tuple otherwise...
        for dependency_command in metadeps.EXTERNAL_COMMANDS.get(
            requirement_name, ())
    )


@type_check
def is_runtime_optional(*requirement_names: str) -> bool:
    '''
    ``True`` only if all optional runtime dependencies of this application with
    the passed :mod:`setuptools`-specific project names are **satisfiable**
    (i.e., both importable and of a satisfactory version) *and* all external
    commands required by these dependencies (e.g., GraphViz's ``dot`` command)
    reside in the current ``${PATH}``.

    Parameters
    ----------
    requirement_names : tuple[str]
        Tuple of the names of all :mod:`setuptools`-specific projects
        implementing these dependencies (e.g., ``NetworkX``). If any such
        name is *not* a key of the :data:`metadeps.RUNTIME_OPTIONAL` dictionary
        and is thus unrecognized, an exception is raised.

    Returns
    ----------
    bool
        ``True`` only if these optional runtime dependencies are satisfiable.

    See Also
    ----------
    :func:`die_unless_runtime_mandatory_all`
        Further details.
    '''

    return is_requirements_dict_keys(
        metadeps.RUNTIME_OPTIONAL, *requirement_names)


@type_check
def is_requirements_dict_keys(
    requirements_dict: MappingType, *requirement_names: str) -> bool:
    '''
    ``True`` only if all dependencies with the passed
    :mod:`setuptools`-specific project names are **satisfiable** (i.e., both
    importable and of a satisfactory version) *and* all external commands
    required by these dependencies (e.g., GraphViz's ``dot`` command) reside in
    the current ``${PATH}``.

    Parameters
    ----------
    requirements_dict : MappingType
        Dictionary mapping from the names of all :mod:`setuptools`-specific
        projects implementing these dependencies to the requirements strings
        constraining these dependencies.
    requirement_names : tuple[str]
        Tuple of the names of all :mod:`setuptools`-specific projects
        implementing these dependencies (e.g., ``NetworkX``). If any such
        name is *not* a key of this dictionary, an exception is raised.

    Returns
    ----------
    bool
        ``True`` only if these dependencies are all satisfiable.

    See Also
    ----------
    :func:`die_unless_runtime_mandatory_all`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.lib.setuptools import setuptool

    # Tuple of requirements strings converted from this subset of key-value
    # pairs of these requirements.
    requirements_tuple = setuptool.convert_requirements_dict_keys_to_tuple(
        requirements_dict, *requirement_names)

    # Return True only if...
    return (
        # These dependencies are all satisfied.
        setuptool.is_requirement_str(*requirements_tuple) and
        # All external commands required by these dependencies are installed.
        is_commands(*requirement_names)
    )

# ....................{ GETTERS ~ runtime                 }....................
def get_runtime_mandatory_tuple() -> tuple:
    '''
    Tuple listing the :mod:`setuptools`-specific requirement string containing
    the mandatory name and optional version and extras constraints of each
    mandatory runtime dependency for this application, dynamically converted
    from the :data:`metadata.RUNTIME_MANDATORY` dictionary.
    '''

    # Avoid circular import dependencies.
    from betse.lib.setuptools import setuptool

    # Convert this dictionary into a tuple.
    return setuptool.convert_requirements_dict_to_tuple(
        metadeps.RUNTIME_MANDATORY)


def get_runtime_optional_tuple() -> tuple:
    '''
    Tuple listing the :mod:`setuptools`-specific requirement string containing
    the mandatory name and optional version and extras constraints of each
    optional runtime dependency for this application, dynamically converted
    from the :data:`metadata.RUNTIME_OPTIONAL` dictionary.
    '''

    # Avoid circular import dependencies.
    from betse.lib.setuptools import setuptool

    # Convert this dictionary into a tuple.
    return setuptool.convert_requirements_dict_to_tuple(
        metadeps.RUNTIME_OPTIONAL)

# ....................{ GETTERS ~ testing                 }....................
def get_testing_mandatory_tuple() -> tuple:
    '''
    Tuple listing the :mod:`setuptools`-specific requirement string containing
    the mandatory name and optional version and extras constraints of each
    mandatory testing dependency for this application, dynamically converted
    from the :data:`metadata.RUNTIME_OPTIONAL` dictionary.
    '''

    # Avoid circular import dependencies.
    from betse.lib.setuptools import setuptool

    # Convert this dictionary into a tuple.
    return setuptool.convert_requirements_dict_to_tuple(
        metadeps.TESTING_MANDATORY)

# ....................{ GETTERS ~ metadata                }....................
def get_metadatas() -> tuple:
    '''
    Tuple of 2-tuples `(metedata_name, metadata_value`), describing all
    currently installed optional and mandatory third-party dependencies
    required at both runtime and testing time.
    '''

    # Defer heavyweight imports.
    from betse.lib.matplotlib.matplotlibs import mpl_config
    from betse.lib.numpy import numpys
    from betse.lib.setuptools import setuptool

    # Tuple of all dependency versions.
    LIB_VERSION_METADATA = (
        # Dependencies metadata.
        ('runtime dependencies (mandatory)',
         setuptool.get_requirements_dict_metadata(metadeps.RUNTIME_MANDATORY)),
        ('runtime dependencies (optional)',
         setuptool.get_requirements_dict_metadata(metadeps.RUNTIME_OPTIONAL)),
        ('testing dependencies (mandatory)',
         setuptool.get_requirements_dict_metadata(metadeps.TESTING_MANDATORY)),
    )

    # Tuple of all dependency-specific metadata.
    LIB_METADATA = (
        # matplotlib metadata.
        ('matplotlib', mpl_config.get_metadata()),
    )

    # Return a tuple aggregating these tuples.
    return (
        LIB_VERSION_METADATA +
        LIB_METADATA +
        numpys.get_metadatas()
    )

# ....................{ IMPORTERS                         }....................
@type_check
def import_runtime_optional(*requirement_names: str) -> object:
    '''
    Import and return the top-level module object satisfying each optional
    runtime dependency of this application with the passed name.

    Parameters
    ----------
    requirement_names : tuple[str]
        Tuple of the names of all :mod:`setuptools`-specific projects
        implementing these dependencies (e.g., ``NetworkX``). If any such name
        is unrecognized (i.e., is *not* a key of the
        :data:`metadeps.RUNTIME_OPTIONAL` dictionary), an exception is raised.

    See Also
    ----------
    :func:`import_requirements_dict_keys`
        Further details.
    '''

    return import_requirements_dict_keys(
        metadeps.RUNTIME_OPTIONAL, *requirement_names)


@type_check
def import_requirements_dict_keys(
    requirements_dict: MappingType, *requirement_names: str) -> object:
    '''
    Import and return the top-level module object satisfying each dependency
    with the passed name described by the passed dictionary.

    See Also
    ----------
    :func:`setuptool.import_requirements_dict_keys`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.lib.setuptools import setuptool

    # Validate these dependencies.
    die_unless_requirements_dict_keys(requirements_dict, *requirement_names)

    # Validate all external commands required by these dependencies.
    return setuptool.import_requirements_dict_keys(
        requirements_dict, *requirement_names)

# ....................{ INITIALIZERS                      }....................
def reinit(*args, **kwargs) -> None:
    '''
    (Re-)initialize all mandatory runtime dependencies of this application with
    the passed parameters.

    Parameters
    ----------
    All passed parameters are passed to the :func:`init` function as is.
    '''

    # Force the init() function to reinitialize this application.
    global _IS_INITTED
    _IS_INITTED = False

    # Reinitialize these dependencies with these parameters.
    init(*args, **kwargs)


@type_check
def init(matplotlib_backend_name: StrOrNoneTypes = None) -> None:
    '''
    Initialize all mandatory runtime dependencies of this application if this
    function has not already been called *or* silently reduce to a noop
    otherwise (i.e., if this function has already been called).

    Specifically, this function (in no particular order):

    * Reconfigures matplotlib with sane defaults specific to the current
      platform and set of all available third-party GUI frameworks.
    * Initializes exactly one available third-party YAML parsing framework
      (e.g., PyYaml, :mod:`ruamel.yaml`).
    * Initializes NumPy.
    * Initializes Pillow.

    Parameters
    ----------
    matplotlib_backend_name : StrOrNoneTypes
        Name of the matplotlib backend to explicitly enable. Defaults to
        ``None``, in which case this method implicitly enables the first
        importable backend known to be both usable and supported by this
        application (in descending order of preference).
    '''

    # If this function has already been called, noop.
    global _IS_INITTED
    if     _IS_INITTED:
        return

    # Defer heavyweight imports.
    from betse.lib.matplotlib.matplotlibs import mpl_config
    from betse.lib.numpy import numpys
    from betse.lib.pickle import pickles
    from betse.lib.pil import pils
    from betse.lib.yaml import yamls

    # Log this initialization. Since initializing heavyweight third-party
    # dependencies (especially matplotlib) consumes non-trivial time, this
    # message is intentionally exposed to all users by default.
    logs.log_info('Loading third-party %s dependencies...', metadata.NAME)

    # Initialize these dependencies.
    mpl_config.init(backend_name=matplotlib_backend_name)
    numpys.init()
    pickles.init()
    pils.init()
    yamls.init()

    # Record this function as having been called *AFTER* successfully doing so.
    _IS_INITTED = True



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

# ....................{ IMPORTS ~ fixture                  }....................
'''
Global functional test configuration for BETSE's command-line interface (CLI).

``py.test`` implicitly imports all functionality defined by this module into all
CLI-specific functional test modules. As this functionality includes all
publicly declared functional fixtures in this ``fixture`` subpackage, these
tests may reference these fixtures without explicit imports.
'''

# ....................{ IMPORTS ~ fixture                  }....................
from betse_test.func.fixture.clier import betse_cli
from betse_test.func.fixture.sim.clisimer import (
    betse_cli_sim,
    betse_cli_sim_default,
    betse_cli_sim_compat,
)
